//======================================================================
// File:		breadboard.cpp
// Author:	Matthias Toussaint
// Created:	Mon Jul 26 23:08:00 CET 2004
//----------------------------------------------------------------------
// Permission to use, copy, modify, and distribute this software and its
// documentation  for any  purpose and  without fee is  hereby  granted,
// provided  that below copyright notice appear  in all copies  and that
// both  that  copyright  notice and  this permission  notice  appear in
// supporting documentation.
// 
// This  file is  provided AS IS  with no  warranties  of any kind.  The
// author shall  have no liability  with respect  to the infringement of
// copyrights, trade  secrets  or any patents by  this file  or any part
// thereof.  In no event will the author be liable  for any lost revenue
// or profits or other special, indirect and consequential damages.
//----------------------------------------------------------------------
// (c) 2004 Matthias Toussaint
//======================================================================

#include <qpainter.h>

#include <breadboard.h>
#include <pic.h>
#include <sevensegmentdisplay.h>
#include <matrixdisplay_7x5.h>
#include <pic16f628.h>

#include <iostream>

BreadBoard::BreadBoard( QWidget *parent, const char *name ) :
  QScrollView( parent, name ),
  m_selectedPart( 0 ),
  m_mode( Selection ),
  m_drawConnection( false )
{
  resizeContents( 1024, 1024 );
  viewport()->setBackgroundMode( QWidget::NoBackground );
}

BreadBoard::~BreadBoard()
{
}

void
BreadBoard::show()
{
  QScrollView::show();
  refreshSLOT();
}

void
BreadBoard::setPic( Pic * pic )
{
  SevenSegmentDisplay *disp = new SevenSegmentDisplay( "Segment A" );
  
  for (int i=0; i<7; ++i)
  {
    disp->assignRegister( i, pic->sfr( Pic::TMR1L ), i );
  }
    
  disp->move( 160, 160 );
  
  m_parts.append( disp );
  
  disp = new SevenSegmentDisplay( "Segment B" );
  
  for (int i=0; i<7; ++i)
  {
    disp->assignRegister( i, pic->sfr( Pic::TMR1L ), i );
  }
    
  disp->move( 320, 160 );
  
  m_parts.append( disp );
  
  disp = new SevenSegmentDisplay( "Segment C" );
  
  for (int i=0; i<7; ++i)
  {
    disp->assignRegister( i, pic->sfr( Pic::TMR1L ), i );
  }
    
  disp->move( 480, 160 );
  
  m_parts.append( disp );
  
  Pic16F628 *picPart = new Pic16F628;
  
  m_parts.append( picPart );
  
  picPart = new Pic16F628;
  
  m_parts.append( picPart );
  
  picPart = new Pic16F628;
  
  m_parts.append( picPart );
  
  m_parts.append( matrixDisplay( pic ) );
  
  refreshSLOT();
}

MatrixDisplay_7x5 *
BreadBoard::matrixDisplay( Pic * pic )
{  
  MatrixDisplay_7x5 *mat = new MatrixDisplay_7x5();
  
  for (int x=0; x<5; ++x)
  {
    mat->assignXRegister( x, pic->sfr( Pic::TMR1H ), x );
  }
  
  for (int y=0; y<7; ++y)
  {
    mat->assignYRegister( y, pic->sfr( Pic::TMR1L ), y );
  }
  
  return mat;
}

void
BreadBoard::drawContents( QPainter *p, int x, int y, int w, int h )
{
  QRect clipRect( x, y, w, h ); 
  QRegion region( clipRect );
  
  PartListIterator it( m_parts );  
    
  for (; *it; ++it)
  {
    QRect rect = (*it)->rect();
    rect.moveBy( -contentsX(), -contentsY() );
    region -= rect; 
  }
  
  clipRect.moveBy( contentsX(), contentsY() );
  
  for (it.toFirst(); *it; ++it)
  {
    if (clipRect.intersects( (*it)->rect() ))
    {
      p->save();
      p->translate( (*it)->rect().x(), 
                    (*it)->rect().y() );
      (*it)->paint( p );
      p->restore();
    }
  }
  
  p->setClipRegion( region );  
  p->fillRect( x, y, w, h, colorGroup().base() );
  
  for (it.toFirst(); *it; ++it)
  {
    if (clipRect.intersects( (*it)->rect() ))
    {
      p->save();
      p->translate( (*it)->rect().x(), 
                    (*it)->rect().y() );
      (*it)->drawHandles( p );
      p->restore();
    }
  }
  
  if (m_drawConnection)
  {
    p->setPen( Qt::blue );
    
    p->moveTo( m_lastMouse.x(), m_lastMouse.y() );
    p->lineTo( m_connectionPos.x(), m_lastMouse.y() );
    p->lineTo( m_connectionPos.x(), m_connectionPos.y() );
  }
  
}

void
BreadBoard::refreshSLOT()
{
  if (!isVisible()) return;
  
  viewport()->update();
}

void
BreadBoard::setModeSLOT( Mode mode )
{
  m_mode = mode;
}

QPoint 
BreadBoard::gridPos( const QPoint & pos ) const
{
  return QPoint( pos.x() - (pos.x() % 16) + 8, 
                 pos.y() - (pos.y() % 16) + 8 );
}

void
BreadBoard::contentsMousePressEvent( QMouseEvent *ev )
{
  QPoint m = gridPos( ev->pos() );  
  m_lastMouse = QPoint( m.x(), m.y() );
  
  if (Selection == mode())
  {
    PartListIterator it( m_parts );  

    m_selectedPart = 0;

    for (it.toFirst(); *it; ++it)
    {
      (*it)->setSelected( false );

      if ((*it)->rect().contains( ev->pos() ))
      {
        (*it)->setSelected( true );
        m_selectedPart = (*it);
      }
    }
  }
  else 
  {
    std::cerr << "hit test ..." << std::endl;
    for (PartListIterator it( m_parts ); *it; ++it)
    {
      std::cerr << "  checking ... " << (*it)->name() << std::endl;
      if ((*it)->hitPin( m_lastMouse ) >= 0)
      {
        std::cerr << "    HIT" << std::endl;
        
        m_drawConnection = true;
        viewport()->update();
        
        break;
      }
    }
  }
  
  viewport()->update();
}

void
BreadBoard::contentsMouseReleaseEvent( QMouseEvent *ev )
{
  QPoint m = gridPos( ev->pos() );  
  m_lastMouse = QPoint( m.x(), m.y() );
  
  if (mode() == CreateConnection)
  {
    m_drawConnection = false;
    viewport()->update();
  }
}

void
BreadBoard::contentsMouseMoveEvent( QMouseEvent *ev )
{
  QPoint m = gridPos( ev->pos() );  
  
  const int dx = m.x() - m_lastMouse.x();
  const int dy = m.y() - m_lastMouse.y();
  
  if (dx || dy)
  {
    m_connectionPos = QPoint( m.x(), m.y() );
    
    if (Selection == mode())
    {
      if (!m_selectedPart) return;
  
      m_selectedPart->move( dx, dy );
      m_lastMouse = m_connectionPos;
      viewport()->update();
    }
    else
    {
      viewport()->update();
    }
  }
}

