//======================================================================
// File:		cod.h
// Author:	Matthias Toussaint
// Created:	Mon Jul 26 23:08:00 CET 2004
//----------------------------------------------------------------------
// Permission to use, copy, modify, and distribute this software and its
// documentation  for any  purpose and  without fee is  hereby  granted,
// provided  that below copyright notice appear  in all copies  and that
// both  that  copyright  notice and  this permission  notice  appear in
// supporting documentation.
// 
// This  file is  provided AS IS  with no  warranties  of any kind.  The
// author shall  have no liability  with respect  to the infringement of
// copyrights, trade  secrets  or any patents by  this file  or any part
// thereof.  In no event will the author be liable  for any lost revenue
// or profits or other special, indirect and consequential damages.
//----------------------------------------------------------------------
// (c) 2004 Matthias Toussaint
//======================================================================

#ifndef COD_HH
#define COD_HH

#include <qstring.h>

class Pic;

/** Base class for 512 byte block from .cod file.
  */
class CodBlock
{
public:
  enum
  {
    Type_ShortSymbols = 4,
    Type_LongSymbols = 5,
    Type_List,
    Type_MemoryMap,
    Type_LocalVariables,
    Type_SourceFiles,
    Type_DebugMessages,
    MaxBlocks                 = 128,
    BlockSize                 = 512
  };

  CodBlock( char * );
  virtual ~CodBlock();
  
  int value( int offset ) const;
  int shortInt( int offset ) const;
  unsigned char byte( int i ) const { return m_data[i]; }
  
  unsigned char *data() const { return m_data; }
  
protected:
  unsigned char *m_data;

};

/** Specialized directory block from .cod file.
  */
class DirBlock : public CodBlock
{
public:
  enum
  {
    Offset_Code            = 0,   
    Offset_Source          = 257, 
    Offset_Date            = 321, 
    Offset_Time            = 328, 
    Offset_Version         = 331, 
    Offset_Compiler        = 351, 
    Offset_Notice          = 363, 
    Offset_SymbolTable     = 426, 
    Offset_NameTable       = 430, 
    Offset_List            = 434, 
    Offset_AddressSize     = 438, 
    Offset_HighAddress     = 439, 
    Offset_NextDirectory   = 441, 
    Offset_MemoryMap       = 443, 
    Offset_LocalVariables  = 447, 
    Offset_CodType         = 451, 

    Offset_Processor       = 454, 
    Offset_LongSymbols     = 462, 
    Offset_DebugMessages   = 466
  }; 

  DirBlock( char * );
  virtual ~DirBlock();

  const char *filename() const { return (char *)&m_data[Offset_Source]; }
  const char *copyright() const { return (char *)&m_data[Offset_Notice]; }
  const char *processor() const { return (char *)&m_data[Offset_Processor]; }
  const char *date() const { return (char *)&m_data[Offset_Date]; }
  const char *time() const { return (char *)&m_data[Offset_Time]; }
  const char *version() const { return (char *)&m_data[Offset_Version]; }
  const char *compiler() const { return (char *)&m_data[Offset_Compiler]; }
  
  /// Retrieve offset for memory map block
  int memoryMap() const { return shortInt( Offset_MemoryMap ); }  
  /// Retrieve offset for code block
  int code() const { return shortInt( Offset_Code ); }
  /// Retrieve offset for source block
  int source() const { return shortInt( Offset_Source ); }
  /// 
  int addressSize() const { return shortInt( Offset_AddressSize ); }
  /// Retrieve offset for next dir block
  int nextDir() const { return shortInt( Offset_NextDirectory ); }
  /// Retrieve offset for local variables (C) block
  int localVariables() const { return shortInt( Offset_LocalVariables ); }
  int codType() const { return shortInt( Offset_CodType ); }
  /// Retrieve offset for long symbols (default for gpasm) block
  int longSymbols() const { return shortInt( Offset_LongSymbols ); }
  /// ???
  int debugMessages() const { return shortInt( Offset_DebugMessages ); }
  /// Retrieve offset for symbol table block
  int symbolTable() const { return shortInt( Offset_SymbolTable ); }
  /// Retrieve offset for filename(s) block
  int nameTable() const { return shortInt( Offset_NameTable ); }

};

class DataBlock : public CodBlock
{
public:
  DataBlock( char * );
  virtual ~DataBlock();

};
 
class NameBlock : public CodBlock
{
public:
  enum
  {
    NameSize = 64,
    NumNames = 8
  };
  
  NameBlock( char * );
  virtual ~NameBlock();
  
  char *name( int i ) const { return (char *)&m_data[i*NameSize+1]; }
  
};
 
class SymbolBlock : public CodBlock
{
public:
  enum
  {
    SymbolLength = 0,
    Offset = 7
  };
  
  SymbolBlock( char * );
  virtual ~SymbolBlock();
  
  int numSymbols() const;
  QString symbol( int i ) const;
  int symbolType( int i ) const;
  int symbolValue( int i ) const;
  int symbolOffset( int i ) const;
};

class Cod
{
public: 
  Cod( const QString &, Pic *pic );
  ~Cod();
  
protected:
  DirBlock   *m_dirBlock;
  DataBlock  *m_block[CodBlock::MaxBlocks];
  int        m_numBlocks;
  
  bool readFile( const QString & );
  
};

inline int
CodBlock::shortInt( int offset ) const
{
  return ((unsigned char)byte(offset) + 
          ((unsigned char)byte(offset+1) << 8));
}

inline int
CodBlock::value( int offset ) const
{
  return ((unsigned char)byte(offset+1) + 
          ((unsigned char)byte(offset) << 8));
}

#endif // COD_HH
