//======================================================================
// File:		cpuview.cpp
// Author:	Matthias Toussaint
// Created:	Mon Jul 26 23:08:00 CET 2004
//----------------------------------------------------------------------
// Permission to use, copy, modify, and distribute this software and its
// documentation  for any  purpose and  without fee is  hereby  granted,
// provided  that below copyright notice appear  in all copies  and that
// both  that  copyright  notice and  this permission  notice  appear in
// supporting documentation.
// 
// This  file is  provided AS IS  with no  warranties  of any kind.  The
// author shall  have no liability  with respect  to the infringement of
// copyrights, trade  secrets  or any patents by  this file  or any part
// thereof.  In no event will the author be liable  for any lost revenue
// or profits or other special, indirect and consequential damages.
//----------------------------------------------------------------------
// (c) 2004 Matthias Toussaint
//======================================================================

#include <qcheckbox.h>
#include <qheader.h>
#include <qlabel.h>
#include <qlineedit.h>
#include <qtable.h>
#include <qvalidator.h>

#include <bittableitem.h>
#include <hextableitem.h>
#include <dectableitem.h>
#include <cpuview.h>
#include <stringconversion.h>

#include <iostream>

CPUView::CPUView( QWidget *parent, const char *name ) :
  UICPUView( parent, name ),
  m_pic( 0 )
{
  connect( ui_table, SIGNAL( valueChanged( int, int ) ),
           this, SLOT( valueChangedSLOT( int, int ) ));
  connect( ui_table, SIGNAL( clicked( int, int, int, const QPoint &) ),
           this, SLOT( clickedSLOT( int, int, int, const QPoint &) ) );
  
  for (int i=0; i<128*4; ++i)
  {
    m_lastValue[i] = -1;
  }
  
  ui_clockSpeed->setValidator( new QDoubleValidator( 0.5, 20.0, 2, this ) );
  
  connect( ui_clockSpeed, SIGNAL( textChanged( const QString & ) ),
           this, SLOT( clockSpeedChangedSLOT( const QString & ) ));
  
  setEditable( true );
}

CPUView::~CPUView()
{
}

void
CPUView::show()
{
  UICPUView::show();
  refreshSLOT();
}

void
CPUView::setEditable( bool /*editable*/ )
{
  m_editable = true;
  //m_editable = editable;
  ui_table->setReadOnly( !m_editable );
}

void
CPUView::setPic( Pic * pic )
{
  m_pic = pic;
  
  if (!m_pic)
  {
    ui_table->setNumRows(0);
    m_registerCnt=0;
    
    return;
  }
  
  PicRAM *ram = m_pic->ram();
  
  m_registerCnt = 0;
  for (int i=0; i<m_pic->ram()->numBanks()*128; ++i)
  {
    if (ram->get(i)->isSFR() && !ram->isShadow(i)) 
    {
      m_register[m_registerCnt++] = ram->get(i);
    }
  }
  
  ui_table->setNumRows( m_registerCnt );
  
  QHeader *vHeader = ui_table->verticalHeader();
  
  for (int i=0; i<m_registerCnt; ++i)
  {
    vHeader->setLabel( i, m_register[i]->name() );

    ui_table->setItem( i, 0, new HexTableItem( ui_table, m_register[i] ) );
    ui_table->setItem( i, 1, new DecTableItem( ui_table, m_register[i] ) );
    
    for (int b=0; b<8; ++b)
    {     
      if (m_register[i]->hasBit(b))
      {
        BitTableItem *item = 
          new BitTableItem( ui_table, m_register[i]->flagName(b) );
        ui_table->setItem( i, 9-b, item );
      }
      else
      {
        QTableItem *item = 
          new QTableItem( ui_table, QTableItem::Never, "" );
        ui_table->setItem( i, 9-b, item );
      }
    }
  }
  
  for (int i=0; i<10; ++i)
  {
    ui_table->adjustColumn( i );
  }
  
  std::cerr << "found " << m_registerCnt << " SFR" << std::endl;
}

void
CPUView::refreshSLOT()
{
  static QString str;
  
  if (!m_pic) return;
  if (!isVisible()) return;
  
  //std::cerr << "refreshSLOT()" << std::endl;
  
  ui_wHex->setText( StringConversion::hexString( m_pic->workingRegister() ) );  
  ui_wDec->setText( StringConversion::decString( m_pic->workingRegister() ) );
  
  str.sprintf( "%03X", m_pic->pc() );
  ui_pc->setText( str );
  
  double freq = m_pic->clockSpeed();

  str.sprintf( tr("(%.2f%% ~ %s)"), 
               m_pic->percentSpeed(), 
               StringConversion::freqString( freq ).latin1() );
  ui_realSpeed->setText( str );
  
  for (int i=0; i<m_registerCnt; ++i)
  {
    if (m_lastValue[i] != m_register[i]->freezedValue())
    {
      updateRow( i );
    }
    
    m_lastValue[i] = m_register[i]->freezedValue();
  }
}

void
CPUView::setValue( int row, int value )
{
  m_register[row]->setValuePic( value );
  m_register[row]->freeze();
  
  updateRow( row );
}

void
CPUView::updateRow( int row )
{
  const int value = m_register[row]->freezedValue();
  
  HexTableItem *hi = (HexTableItem *)ui_table->item( row, 0 ); 
  hi->update();
  ui_table->updateCell( row, 0 );

  DecTableItem *di = (DecTableItem *)ui_table->item( row, 1 ); 
  di->update();
  ui_table->updateCell( row, 1 );

  unsigned bits = 0x80;

  for (int b=0; b<8; ++b)
  {
    BitTableItem *item = 
      dynamic_cast<BitTableItem *> (ui_table->item( row, b+2 ));

    if (item)
    {
      item->setChecked( value & bits );
    }

    bits >>= 1;
  }
}

void
CPUView::valueChangedSLOT( int row, int col )
{
  static QString str;
    
  if (col <= 1)
  {
    const int value = 
        ui_table->text( row, col ).toInt( 0, 0==col ? 16 : 10 ) & 
          m_register[row]->mask();
        
    setValue( row, value );
  }
    
  emit registerChanged( m_register[row] );
}

void
CPUView::clickedSLOT( int row, int col, int /*but*/, const QPoint & )
{   
  static QString str;
    
  if (!m_editable) return;
  
  if (col > 1)
  {
    BitTableItem *item = 
        dynamic_cast<BitTableItem *> (ui_table->item( row, col ));
    
    if (item)
    {
      item->setChecked( !item->isChecked() );

      int value = 0;

      unsigned bits = 0x80;

      for (int b=0; b<8; ++b)
      {
        BitTableItem *item = 
          dynamic_cast<BitTableItem *> (ui_table->item( row, b+2 ));

        if (item && item->isChecked())
        {
          value |= bits;
        }

        bits >>= 1;
      }

      setValue( row, value );
    }
  }
    
  emit registerChanged( m_register[row] );
}

void
CPUView::clockSpeedChangedSLOT( const QString & text )
{
  double const freq = text.toDouble() * 1e6;
  
  if (freq > 1)
  {
    m_pic->setCrystalFrequency( freq );
  }
}
