//======================================================================
// File:		highlightcondition.cpp
// Author:	Matthias Toussaint
// Created:	Mon Jul 26 23:08:00 CET 2004
//----------------------------------------------------------------------
// Permission to use, copy, modify, and distribute this software and its
// documentation  for any  purpose and  without fee is  hereby  granted,
// provided  that below copyright notice appear  in all copies  and that
// both  that  copyright  notice and  this permission  notice  appear in
// supporting documentation.
// 
// This  file is  provided AS IS  with no  warranties  of any kind.  The
// author shall  have no liability  with respect  to the infringement of
// copyrights, trade  secrets  or any patents by  this file  or any part
// thereof.  In no event will the author be liable  for any lost revenue
// or profits or other special, indirect and consequential damages.
//----------------------------------------------------------------------
// (c) 2004 Matthias Toussaint
//======================================================================

#include <highlightcondition.h>
#include <picram.h>

#include <iostream>

HighlightCondition::HighlightCondition() :
  m_ram( 0 ),
  m_compareValue( 0 ),
  m_registerAddr( -1 ),
  m_mode( Unknown ),
  m_patternWasEmpty( true ),
  m_isIncomplete( true )
{
}

HighlightCondition::HighlightCondition( PicRAM *ram ) :
  m_ram( ram ),
  m_compareValue( 0 ),
  m_registerAddr( -1 ),
  m_mode( Unknown ),
  m_patternWasEmpty( true ),
  m_isIncomplete( true )
{
}

HighlightCondition::~HighlightCondition()
{
}

void
HighlightCondition::setRAM( PicRAM *ram )
{
  m_ram = ram;
}

void
HighlightCondition::setPattern( const QString & pattern )
{
  QString tmpPattern = pattern.simplifyWhiteSpace().stripWhiteSpace().upper();
   
  //std::cerr << tmpPattern.latin1() << std::endl;
  
  setPatternWasEmpty( tmpPattern.isEmpty() );
  
  QString command;
  QString compare;
  QString regAddr;
  
  unsigned pos = 0;
  
  // get addr of register (if any)
  //
  while (pos < tmpPattern.length() && 
         !isCondition( tmpPattern.constref(pos) ) &&
         !tmpPattern.constref(pos).isSpace())
  {
    regAddr += tmpPattern.constref(pos);
    ++pos;
  }
  
  skipWhiteSpace( &tmpPattern, &pos );
  
  // get condition
  //
  while (pos < tmpPattern.length() && 
         isCondition( tmpPattern.constref(pos) ))
  {
    command += tmpPattern.constref(pos);
    ++pos;
  }
  
  skipWhiteSpace( &tmpPattern, &pos );
  
  // get condition value
  //
  while (pos < tmpPattern.length() && 
         isHexNumber( tmpPattern.constref(pos) ))
  {
    compare += tmpPattern.constref(pos);
    ++pos;
  }
  
  if (regAddr.isEmpty())
  {
    m_registerAddr = -1;
  }
  else
  {
    m_registerAddr = m_ram->symbolToAddr( regAddr );

    if (m_registerAddr == -1)
    {
      bool ok=false;
      m_registerAddr = regAddr.toInt( &ok, 16 ) & 0x7ff ;

      if (!ok && m_registerAddr >= 0 && m_registerAddr <= 0x7ff)
      {
        m_registerAddr = -1;
      }
    }
  }
  
  m_compareValue = -1;
  
  if (!compare.isEmpty())
  {
    bool ok=false;
    m_compareValue = compare.toInt( &ok, 16 ) & 0xff;
     
    if (!ok)  
    { 
      m_compareValue = -1;
    }
  }
  
  if      (command == "==") setMode( Equal );
  else if (command == "!=") setMode( NotEqual );
  else if (command == ">")  setMode( GreaterThan );
  else if (command == ">=") setMode( GreaterOrEqual );
  else if (command == "<")  setMode( LessThan );
  else if (command == "<=") setMode( LessOrEqual );
  else if (command == "&")  setMode( And );
  else                      setMode( Unknown );
  
  setIncomplete( true );
  
  if (compare.isEmpty() && command.isEmpty() && m_registerAddr != -1)
  {
    //std::cerr << "first case" << std::endl;
    setIncomplete( false );
  }
  
  if (mode() != Unknown && m_compareValue != -1)
  {
    //std::cerr << "second case" << std::endl;
    setIncomplete( false );
  }
}

bool
HighlightCondition::syntaxOk() const
{
  if (incomplete())
  {
    if (patternWasEmpty()) return true;
  }
  
  return !incomplete();
}

bool
HighlightCondition::match( int addr, int value ) const
{
  if (incomplete()) return false;
  
  if (m_registerAddr != -1 && addr != m_registerAddr)
  {
    return false;
  }
  
  switch (m_mode)
  {
  case GreaterThan:
    return (value > m_compareValue);
    
  case GreaterOrEqual:
    return (value >= m_compareValue);
    
  case LessThan:
    return (value < m_compareValue);
    
  case LessOrEqual:
    return (value <= m_compareValue);
    
  case Equal:
    return (value == m_compareValue);
    
  case NotEqual:
    return (value != m_compareValue);
    
  case And:
    return (value & m_compareValue);
    
  case Unknown:
    return (m_registerAddr != -1);
  }
  
  return false;
}
    
