//======================================================================
// File:		instructiontraceviewtable.cpp
// Author:	Matthias Toussaint
// Created:	Mon Jul 26 23:08:00 CET 2004
//----------------------------------------------------------------------
// Permission to use, copy, modify, and distribute this software and its
// documentation  for any  purpose and  without fee is  hereby  granted,
// provided  that below copyright notice appear  in all copies  and that
// both  that  copyright  notice and  this permission  notice  appear in
// supporting documentation.
// 
// This  file is  provided AS IS  with no  warranties  of any kind.  The
// author shall  have no liability  with respect  to the infringement of
// copyrights, trade  secrets  or any patents by  this file  or any part
// thereof.  In no event will the author be liable  for any lost revenue
// or profits or other special, indirect and consequential damages.
//----------------------------------------------------------------------
// (c) 2004 Matthias Toussaint
//======================================================================

#include <qpainter.h>
#include <qscrollbar.h>
#include <qheader.h>

#include <instructiontraceviewtable.h>
#include <instructiontracebuffer.h>
#include <pic.h>
#include <stringconversion.h>

InstructionTraceViewTable::InstructionTraceViewTable( QWidget *parent, const char *name ) :
  QScrollView( parent, name ),
  m_current( 0 )
{
  viewport()->setBackgroundMode( QWidget::NoBackground );
  
  setVScrollBarMode( QScrollView::AlwaysOn );
  setHScrollBarMode( QScrollView::AlwaysOff );
  
  m_lh = fontMetrics().height()+2;
  
  m_header = new QHeader( this );
  m_header->move( 2, 2 );
  setMargins( 0, m_header->height(), 0, 0 );
  
  m_header->addLabel( tr("Count") );
  m_header->addLabel( tr("PC") );
  m_header->addLabel( tr("Instruction") );
  m_header->addLabel( tr("W") );
  m_header->addLabel( tr("File") );
  m_header->addLabel( tr("Status") );
  
  m_header->setTracking( true );
  m_header->setMovingEnabled( false );
  
  connect( m_header, SIGNAL( sizeChange(int, int, int) ),
           this, SLOT( refreshSLOT() ));

  QFont tmpFont = font();
  setFont( QFont( "Fixed", 10 ) );
  m_header->setFont( tmpFont );
  
  QFontMetrics fm = fontMetrics();
  m_header->resizeSection( 1, 
                           fm.width( "000" ) + 8 );
  m_header->resizeSection( 2, 
                           fm.width( "addlw  0xa0 (160)" ) + 8 );
  m_header->resizeSection( 3, 
                           fm.width( "00 (000)" ) + 8 );
  m_header->resizeSection( 4, 
                           fm.width( "0x00=00" ) + 8 );
  m_header->resizeSection( 5, 
                           QMAX( m_header->sectionSize(5), 
                                 fm.width( "C DC Z" ) + 8 ));
}

InstructionTraceViewTable::~InstructionTraceViewTable()
{
}

void
InstructionTraceViewTable::resizeEvent( QResizeEvent * )
{
  refreshSLOT();
}

void
InstructionTraceViewTable::setPic( Pic *pic )
{
  m_pic = pic;
}

void
InstructionTraceViewTable::refreshSLOT()
{
  m_header->resize( QMIN( visibleWidth(), m_header->headerWidth() ), 
                    fontMetrics().height()+8 );
  setMargins( 0, m_header->height(), 0, 0 );
  
  resizeContents( visibleWidth(), 
                  QMAX( visibleHeight(), 
                        m_pic->instructionTraceBuffer()->length()*m_lh ) );
  
  viewport()->update();
}

void 
InstructionTraceViewTable::drawContents( QPainter *p, 
                              int /*cx*/, int cy, int /*cw*/, int ch )
{
  const int startLine = cy / m_lh;
  const int lastLine = (cy+ch) / m_lh + 1;
  const QColor bg = colorGroup().base();
  const QColor bgDark = colorGroup().base().dark( 105 );
  const QColor darkPen = colorGroup().dark();
  const QColor textPen = colorGroup().text();
  const int w = contentsWidth();
  
  QString str;
  
  InstructionTraceBuffer *buffer = m_pic->instructionTraceBuffer();
  PicROM *rom = m_pic->rom();
  
  for (int i=startLine; i<lastLine; ++i)
  {
    const int y = i*m_lh;
    
    if (i & 1)
    {
      p->fillRect( 0, y, w, m_lh, bgDark );
    }
    else
    {
      p->fillRect( 0, y, w, m_lh, bg );
    }
    
    p->setPen( darkPen );
    
    for (int j=0; j<6; ++j)
    {
      p->drawLine( m_header->sectionPos( j )-1, y, 
                   m_header->sectionPos( j )-1, y+m_lh );
    }
    
    p->setPen( textPen );
    
    if (i < m_pic->instructionTraceBuffer()->length())
    {
      const int pc = buffer->pc( i );
      const int opcode = rom->at( pc );
      
      str.sprintf( "%d", i );
      p->drawText( m_header->sectionPos( 0 )+4, y, 
                   m_header->sectionSize( 0 )-8, 
                   m_lh, Qt::AlignLeft | Qt::AlignVCenter, str );
      
      str.sprintf( "%03X", pc );
      p->drawText( m_header->sectionPos( 1 )+4, y, 
                   m_header->sectionSize( 1 )-8, 
                   m_lh, Qt::AlignLeft | Qt::AlignVCenter, str );
      
      p->drawText( m_header->sectionPos( 2 )+4, y, 
                   m_header->sectionSize( 2 )-8, 
                   m_lh, 
                   Qt::AlignLeft | Qt::AlignVCenter,
                   m_pic->opcodeToString( opcode ));
      
      str = StringConversion::hexString( buffer->w(i) );
      str += "(";
      str += StringConversion::decString( buffer->w(i) );
      str += ")";
      p->drawText( m_header->sectionPos( 3 )+4, y, 
                   m_header->sectionSize( 3 )-8, 
                   m_lh, 
                   Qt::AlignLeft | Qt::AlignVCenter,
                   str );
      
      if (buffer->hasFile(i))
      {
        str = "0x";
        str += StringConversion::hexString( m_pic->file( opcode ) );
        str += "=";
        str += StringConversion::hexString( buffer->file(i) );
        
        p->drawText( m_header->sectionPos( 4 )+4, y, 
                     m_header->sectionSize( 4 )-8, 
                     m_lh, 
                     Qt::AlignLeft | Qt::AlignVCenter,
                     str );
      }
      
      const int status = buffer->status(i);

      str.sprintf( "%c %s %c",
                   status & 0x01 ? 'C' : ' ',
                   status & 0x02 ? "DC" : "  ",
                   status & 0x04 ? 'Z' : ' ' );
   
      p->drawText( m_header->sectionPos( 5 )+4, y, 
                   m_header->sectionSize( 5 )-8, 
                   m_lh, Qt::AlignLeft | Qt::AlignVCenter,
                   str );
      
      if (i==m_current)
      {
        p->drawRect( 0, y, 
                     m_header->sectionPos( 5 )+m_header->sectionSize( 5 ),
                     m_lh );
      }
    }
  }
}

void 
InstructionTraceViewTable::contentsMousePressEvent( QMouseEvent *ev )
{
  setCurrentFromPos( ev->pos().y() );
}

void 
InstructionTraceViewTable::contentsMouseMoveEvent( QMouseEvent *ev )
{
  if (m_current != ev->pos().y() / m_lh )
  {
    setCurrentFromPos( ev->pos().y() );
  }
}

void 
InstructionTraceViewTable::contentsMouseDoubleClickEvent( QMouseEvent *ev )
{
  setCurrentFromPos( ev->pos().y() );
  
  emit showPc( m_pic->instructionTraceBuffer()->pc( m_current ) );
}

void 
InstructionTraceViewTable::setCurrentFromPos( int y )
{
  m_current = y / m_lh;
  viewport()->update();
}

