//======================================================================
// File:		pic.cpp
// Author:	Matthias Toussaint
// Created:	Mon Jul 26 23:08:00 CET 2004
//----------------------------------------------------------------------
// Permission to use, copy, modify, and distribute this software and its
// documentation  for any  purpose and  without fee is  hereby  granted,
// provided  that below copyright notice appear  in all copies  and that
// both  that  copyright  notice and  this permission  notice  appear in
// supporting documentation.
// 
// This  file is  provided AS IS  with no  warranties  of any kind.  The
// author shall  have no liability  with respect  to the infringement of
// copyrights, trade  secrets  or any patents by  this file  or any part
// thereof.  In no event will the author be liable  for any lost revenue
// or profits or other special, indirect and consequential damages.
//----------------------------------------------------------------------
// (c) 2004 Matthias Toussaint
//======================================================================

#include <pic.h>
#include <picregister.h>
#include <tools.h>
#include <pictimer0.h>
#include <pictimer16.h>

#include <stdio.h>
#include <math.h>

Pic::Pic() : 
  QThread(),
  m_w( 0 ),
  m_pc( 0 ),
  m_pipelinePos( 0 ),
  m_cachedClockSpeed( 0.0 ),
  m_tracing( false ),
  m_cycles( 0 ),
  m_numOpcode( 37 ),
  m_lastCycles( 0 ),
  m_run( false ),
  m_step( false ),
  m_over( false ),
  m_finishCall( false ),
  m_cont( false ),
  m_rst( false ),
  m_quit( false ),
  m_freeRun( false ),
  m_dirty( true ),
  m_animating( false ),
  m_overBreakpointPc( 0 ),
  m_callDepth( 0 )
{
  createCodeTable();
    
  m_rom = new PicROM( this );
  createRAM();
  createTimer();
    
  m_symbolTable = new SymbolTable;
  
  m_status = m_sfr[STATUS];
  m_pcl    = m_sfr[PCL];
  m_pclath = m_sfr[PCLATH];
  
  m_disassembler = new DisAss;
  m_stack = new PicStack;
  
  m_instructionTraceBuffer = new InstructionTraceBuffer;
  m_registerTrace = new RegisterTrace( this );
  
  resetExec();
  
  setCrystalFrequency( 4e6 );
  
  m_startTime.start();
  
  setHoldSpeed( true );
  
  // fill with nop's
  //
  compile();
  m_rom->reset();
}

Pic::~Pic()
{
  delete [] m_codeTable;
  
  delete m_rom;
  delete m_ram;
  delete m_stack;
  delete m_instructionTraceBuffer;
}

void
Pic::lock()
{
  m_mutex.lock();
}

void
Pic::unlock()
{
  m_mutex.unlock();
}

void
Pic::freeze()
{
  //m_mutex.lock();
  m_ram->freeze();
  //m_mutex.unlock();
}

bool 
Pic::stepRequest() const 
{ 
  return m_step; 
}

bool 
Pic::overRequest() const 
{ 
  return m_over; 
}

bool 
Pic::finishCallRequest() const 
{ 
  return m_finishCall; 
}

bool 
Pic::contRequest() const 
{ 
  return m_cont; 
}

bool 
Pic::rstRequest() const 
{ 
  return m_rst; 
}

bool 
Pic::runRequest() const 
{ 
  return m_run; 
}

bool 
Pic::quitRequest() const 
{ 
  return m_quit; 
}

void 
Pic::setStep( bool on ) 
{ 
  m_mutex.lock();
  m_step = on;
  wakeUp();
  m_mutex.unlock();
}

void 
Pic::setOver( bool on ) 
{ 
  m_mutex.lock();
  m_over = on;
  wakeUp();
  m_mutex.unlock();
}

void 
Pic::setFinishCall( bool on ) 
{ 
  m_mutex.lock();
  m_finishCall = on;
  wakeUp();
  m_mutex.unlock();
}

void 
Pic::setCont( bool on ) 
{ 
  m_mutex.lock();
  m_cont = on;
  wakeUp();
  m_mutex.unlock();
}

void 
Pic::setRst( bool on ) 
{ 
  m_mutex.lock();
  m_rst = on;
  wakeUp();
  m_mutex.unlock();
}

void 
Pic::setRun( bool on ) 
{ 
  m_mutex.lock();
  m_run = on;
  wakeUp();
  m_mutex.unlock();
}

void 
Pic::setQuit( bool on ) 
{ 
  m_mutex.lock();
  m_quit = on;
  wakeUp();
  m_mutex.unlock();
}

bool 
Pic::halted() const 
{ 
  return !m_run; 
}

void
Pic::run()
{
  std::cerr << "run ... ";

  while (1)
  {
    m_command.wait();
    
    m_mutex.lock();
    
    if (m_step) 
    { 
      m_step = false;
      //m_pc = pcFromPCL();
      
      m_run = true;
      m_mutex.unlock();
      // may have to skip two cycle instructions
      if (!stepExec()) stepExec();
      // skip next nopinternal to adjust pc
      if (nextIsNopInternal()) stepExec();
      m_run = false;
    }
    else if (m_over) 
    { 
      m_over = false;
      
      // if next instruction is a call
      //
      if (m_instructionPipeline[(m_pipelinePos ^ 1)] == &Pic::CALL)
      {
        // set a temporary breakpoint if there wasn't already one
        //
        m_overBreakpointPc = m_pc+1;

        if (!m_rom->breakpoint( m_overBreakpointPc ))
        {
          m_rom->setBreakpoint( m_overBreakpointPc );    
          Breakpoint *bp = m_rom->breakpoint( m_overBreakpointPc );
          bp->setTemporary();        
        }

        m_run = true;
        m_mutex.unlock();

        contExec(); 
      }
    }
    else if (m_finishCall) 
    { 
      m_finishCall = false;
      
      // set a temporary breakpoint if there wasn't already one
      //
      m_overBreakpointPc = m_stack->top();

      if (!m_rom->breakpoint( m_overBreakpointPc ))
      {
        m_rom->setBreakpoint( m_overBreakpointPc );    
        Breakpoint *bp = m_rom->breakpoint( m_overBreakpointPc );
        bp->setTemporary();        
      }

      m_run = true;
      m_mutex.unlock();

      contExec(); 
    }
    else if (m_cont) 
    { 
      m_cont = false;
      m_run = true;
      m_mutex.unlock();
      
      contExec(); 
    }
    else if (m_rst)  
    {
      m_rst = false;
      m_mutex.unlock();
      
      resetExec(); 
    }
    else if (m_quit)  
    {
      m_mutex.unlock();
      std::cerr << "thread quit !" << std::endl;
      return;
    }
  }
  
  std::cerr << "exit" << std::endl;
}

double
Pic::clockSpeed()
{
  const int elapsed = m_startTime.elapsed();
  
  if (elapsed>1000)
  {
    const long deltaCycles = m_cycles-m_lastCycles;
    
    if (deltaCycles != 0)
    {
      const double crystal = (double)deltaCycles / elapsed * 4000.0;

      m_lastCycles = m_cycles;   
      m_cachedClockSpeed = crystal;

      m_startTime.start();
    }
  }
  
  return m_cachedClockSpeed;
}
    
void
Pic::quit()
{
  if (m_run) halt();
  setQuit( true );
}
    
void
Pic::reset()
{
  setRst( true );
}
    
void
Pic::resetExec()
{
  m_rom->reset();
  m_ram->reset();
  m_instructionTraceBuffer->reset();
  
  m_pc = -1;
  
  flushInstructionPipeline();

  for (int i=0; i<4096; ++i)
  {
    m_callCyclesSum[i]  = 0.0;
    m_callCount[i]      = 0;
    m_callCyclesLast[i] = 0;
    m_callCyclesMin[i]  = 0xffffffff;
    m_callCyclesMax[i]  = 0;
  }
  
  // prefetch first instruction
  //
  stepExec();

  // we don't want to call the cycle from above
  //
  m_cycles = 0;
  m_lastCycles = 0;
  m_callDepth = 0;
  
  m_startTime.start();
}

void
Pic::step()
{
  setStep( true );
}

void
Pic::over()
{
  setOver( true );
}

void
Pic::finishCall()
{
  setFinishCall( true );
}

void
Pic::halt()
{
  setRun( false );
}

void
Pic::cont()
{
  setCont( true );
}

bool
Pic::stepExec()
{
  for (int i=0; i<m_numTimer; ++i)
  {
    m_timer[i]->clockTick();
  }
  
  //std::cerr << "stepExec PC=" << m_pc << std::endl;
  
  // remember PC for tracing
  //
  m_tracePc = m_pc;
  
  // fetch next instruction
  //
  // we increase pc before executing the instruction
  // this way manipulations of PCL are taken into account
  //
  incPc();
  fetchInstruction( m_pc );  
  
  // execute and return false if instruction was internal NOP
  // a single step can be repeated then ...
  //
  return executeInstruction();
}

void
Pic::contExec()
{
  static int msecToWait = 0;
  static int minWait = 0;
 
  // calibration loop
  //
  if (minWait == 0)
  {
    for (int i=0; i<32; ++i)
    {
      m_holdTime.start();
      msleep(1);
      minWait += m_holdTime.elapsed();
    }

    minWait >>= 5;
  }
  
  m_holdTime.start();
  m_holdCycles = m_cycles;
    
  if (m_pc == -1)
  {
    stepExec();
  }
  
  while (runRequest())
  {
    m_mutex.lock();
    
    const int numCycles = 
        static_cast<const int> ( QMAX( 1.0, crystalFrequency() / 4000.0 ) );
    
    for (int i=0; i<numCycles; ++i)
    {
      if (m_rom->isActiveBreakpoint( m_pc ) || PicRegister::watchActive)
      {
        // if it was a temporary breakpoint from step over -> remove
        //
        if (m_rom->isActiveBreakpoint( m_pc ))
        {
          if (m_rom->breakpoint( m_pc )->isTemporary())
          {
            m_rom->removeBreakpoint( m_pc );
          }
        }
        
        m_mutex.unlock();
        setRun( false );
        
        PicRegister::watchActive = false;
        return;
      }    
        
      stepExec();  
    }
    
    m_mutex.unlock();

    // if user halted -> remove temporary breakpoint from
    // step over request (if any)
    //
    if (quitRequest()) 
    {
      Breakpoint *bp = m_rom->breakpoint( m_overBreakpointPc );
      if (bp && bp->isTemporary())
      {
        m_rom->removeBreakpoint( m_overBreakpointPc );
      }
      return;
    }
    
    // if not free running -> wait
    //
    if (!m_freeRun)
    {
      int dt = timeDifference()+msecToWait;

      if (dt > minWait)
      {
        if (dt > 1000) dt = 1000;
        
        do
        {
          m_holdTime.start();
          msleep( 1 );
          dt -= m_holdTime.elapsed();
        }
        while (dt > minWait);
          
        msecToWait = dt;        
        m_holdTime.start();
        m_holdCycles = m_cycles;
      }
    }
  }
  
  std::cerr << "finished" << std::endl;
}

void
Pic::clearCycles()
{
  m_mutex.lock();
  m_cycles = m_holdCycles = m_lastCycles = 0; 
  m_holdTime.start();
  m_startTime.start();
  m_mutex.unlock();
}
 
int
Pic::timeDifference() const
{
  return (int)(4000.0 * ((double)(m_cycles-m_holdCycles) / m_crystalFrequency ) - 
                m_holdTime.elapsed());
}

void
Pic::loadROM( unsigned *data, int count )
{
  m_rom->setData( data, count );
  m_rom->reset();
  
  // precompile
  //
  compile();
  m_rom->reset();
}

/////////////////////////////////////////////////////////
// byte oriented file register operations
/////////////////////////////////////////////////////////

/* ADDWF f,d - Add W and f

   Add the contents ot the W register with register
   'f'. If 'd' is 0 the result is stored in the W
   register. If 'd' is 1 the result is stored back
   in register 'f'
   
   STATUS: C,DC,Z
   CYCLES: 1
   
   0 <= f <= 127
   d [0,1]
   
   (W) + (f) -> (dest)
*/
void 
Pic::ADDWF( unsigned opcode, int operand )
{
  PicRegister *cell = m_ram->getRegister( operand );
  const int value = m_w + cell->value();
  
  setC_Z_DC( cell->valuePic(), value, m_w );
  
  storeResult( value, cell, opcode );
  
  traceInstruction( cell->valuePic() );
}

/* ANDWF f,d - AND W with f

   AND the W register with register 'f'. If 'd' is 0
   the result is stored in the W register. If 'd' is
   1 the result is stored back in register 'f'
   
   STATUS: Z
   CYCLES: 1
   
   0 <= f <= 127
   d [0,1]
   
   (W) .AND. (f) -> (dest)
*/
void
Pic::ANDWF( unsigned opcode, int operand )
{
  PicRegister *cell = m_ram->getRegister( operand );
  const int value = m_w & cell->value();
    
  setZ( value );
  
  storeResult( value, cell, opcode );
   
  traceInstruction( cell->valuePic() );
}

/* CLRF f - Clear f

   The contents of register 'f' are cleared and the
   Z bit is set.
   
   STATUS: Z
   CYCLES: 1
   
   00h -> (f)
   1 -> Z
*/
void
Pic::CLRF( unsigned, int operand )
{
  PicRegister *cell = m_ram->getRegister( operand );
  // user write
  cell->setValue( 0 );
  
  m_status->setValuePic( m_status->valuePic() | Z_BIT );
  //m_status->setBitPic( Z );
  
  traceInstruction( cell->valuePic() );
}

/* CLRW - Clear W

   W register is cleared. Zero bit Z is set.
   
   STATUS: Z
   CYCLES: 1
   
   00h -> (W)
   1 -> Z
*/
void 
Pic::CLRW( unsigned, int  )
{
  m_w = 0;
  
  m_status->setValuePic( m_status->valuePic() | Z_BIT );
  //m_status->setBitPic( Z );
  
  traceInstruction();
}

/* COMF f,d - Complement f

   The contents of register 'f' are complemented.
   If 'd' is 0 the result is stored in W. If 'd'
   is 1 the result is stored back in register 'f'.
   
   STATUS: Z
   CYCLES: 1
  
   0 <= f <= 127
   d [0,1]
    
   (/f) -> (dest)
*/
void 
Pic::COMF( unsigned opcode, int operand )
{
  PicRegister *cell = m_ram->getRegister( operand );
  const int value = cell->value() ^ 0xff;
  
  setZ( value );
  
  storeResult( value, cell, opcode );
  
  traceInstruction( cell->valuePic() );
}

/* DECF f,d - Decrement f

   Decrement register 'f'. If 'd' is 0 the
   result is stored in the W register. If 'd' is
   1 the result is stored back in register 'f'.
   
   STATUS: Z
   CYCLES: 1
   
   0 <= f <= 127
   d [0,1]
   
   (f)-1 -> (dest)
*/ 
void 
Pic::DECF( unsigned opcode, int operand )
{
  PicRegister *cell = m_ram->getRegister( operand );
  const int value = cell->value() - 1; 
  
  setZ( value );
  
  storeResult( value, cell, opcode );
 
  traceInstruction( cell->valuePic() );
}

/* DECFSZ f,d - Decrement f, Skip if 0

   The content of register 'f' are decremented.
   If 'd' is 0 the result is placed in the W register.
   If 'd' is 1 the result is placed back in register
   'f'.
   If the result is 0, the next instruction, which is 
   already fetched, is discarded. A NOP is executed
   instead making it a two-cycle instruction.
   
   STATUS: none
   CYCLES: 1(2)
   
   0 <= f <= 127
   d [0,1]
   
   (f)-1 -> (dest); skip if result = 0
*/
void 
Pic::DECFSZ( unsigned opcode, int operand )
{
  PicRegister *cell = m_ram->getRegister( operand );
  const int value = cell->value() - 1;
  
  setZ( value );
  
  //const int pcAtCall = m_pc;
  //const unsigned cyclesAtCall = m_cycles;
  
  storeResult( value, cell, opcode );
  
  traceInstruction( cell->valuePic() );
  
  if (zeroFlag())
  {
    // this skips the next instruction
    //
    flushInstructionPipeline();
  }
}

/* INCF f,d - Increment f

   The contents of register 'f' are incremented.
   If 'd' is 0 the result is placed in the W register.
   If 'd' is 1 the result is placed back in register
   'f'
   
   STATUS: Z
   CYCLES: 1
   
   0 <= f <= 127
   d [0,1]
   
   (f)+1 -> (dest)
*/
void 
Pic::INCF( unsigned opcode, int operand )
{
  PicRegister *cell = m_ram->getRegister( operand );
  const int value = cell->value() + 1; 
  
  setZ( value );
  
  storeResult( value, cell, opcode );
  
  traceInstruction( cell->valuePic() );
}

/* INCFSZ f,d - Increment f, Skip if 0

   The contents of register 'f' are incremented.
   If 'd' is 0 the result is placed in the W
   register. If 'd' is 1 the result is placed back
   in register 'f'.
   If the result is 0, the next instruction, 
   which is already fetched, is discarded. A NOP is
   executed instead making it a two-cycle
   instruction.

   STATUS: none
   CYCLES: 1(2)
   
   0 <= f <= 127
   d [0,1]
   
   (f)+1 -> (dest), skip if result = 0
*/
void 
Pic::INCFSZ( unsigned opcode, int operand )
{
  PicRegister *cell = m_ram->getRegister( operand );
  const int value = cell->value() + 1;
  
  //const int pcAtCall = m_pc;
  //const unsigned cyclesAtCall = m_cycles;
    
  storeResult( value, cell, opcode );
  
  traceInstruction( cell->valuePic() );
  
  if ((value & 0xff) == 0)
  {
    // this skips the next instruction
    //
    flushInstructionPipeline();
  }
}

/* IORWF f,d -  Inclusive OR W with f

   Inclusive OR the W register with register 'f'.
   If 'd' is 0 the result is placed in the W register.
   If 'd' is 1 the result is placed back in register
   'f'
   
   STATUS: Z
   CYCLES: 1
   
   0 <= f <= 127
   d [0,1]
   
   (W) .OR. (f) -> (dest)
*/
void 
Pic::IORWF( unsigned opcode, int operand )
{
  PicRegister *cell = m_ram->getRegister( operand );
  const int value = m_w | cell->value();
  
  setZ( value );
  
  storeResult( value, cell, opcode );
  
  traceInstruction( cell->valuePic() );
}

/* MOVF f,d - Move f

   The contents of register F is moved to a destination
   dependent upon the status of d. If d = 0,
   destination is W register. If d = 1, the destination
   is file register f itself. d = 1 is usefull to test
   a file register since status flag Z is affected.
   
   STATUS: Z
   CYCLES: 1
   
   0 <= f <= 127
   d [0,1]
   
   (f) -> (dest)
*/
void 
Pic::MOVF( unsigned opcode, int operand )
{
  PicRegister *cell = m_ram->getRegister( operand );
  const int value = cell->value();
  
  setZ( value );  
  
  storeResult( value, cell, opcode );
  
  traceInstruction( cell->valuePic() );
}

/* MOVWF f - Move W to f

   Move data from W register to register 'f'
   
   STATUS: none
   CYCLES: 1
   
   0 <= f <= 127
   
   (W) -> (f)
*/
void 
Pic::MOVWF( unsigned, int operand )
{
  PicRegister *cell = m_ram->getRegister( operand );
  
  // user write
  //
  cell->setValue( m_w );
  
  traceInstruction( cell->valuePic() );
}

/* NOP - No Operation
 
   STATUS: none
   CYCLES: 1
*/
void 
Pic::NOP( unsigned, int  )
{
  traceInstruction();
}

void 
Pic::NOP_Internal( unsigned, int  )
{
}

/* RLF f,d - Rotate left through Carry

   The contents of register 'f' are rotated one bit
   to the left through the Carry Flag. If 'd' is 0 
   the result is placed in the W register. If 'd' is
   1 the result is stored back in register 'f'.
   
   STATUS: C
   CYCLES: 1
   
   0 <= f <= 127
   d [0,1]
   
    <--[C]<--[REGISTER F]<--
   |                        |
    ------------------------
*/
void 
Pic::RLF( unsigned opcode, int operand )
{
  PicRegister *cell = m_ram->getRegister( operand );
  int value = cell->value() << 1;
  
  if (carryFlag())
  {
    value |= 0x01;
  }
  
  setC( value );
  
  storeResult( value, cell, opcode );
  
  traceInstruction( cell->valuePic() );
}

/* RRF f,d - Rotate right through Carry

   The contents of register 'f' are rotated one bit
   to the right through the Carry Flag. If 'd' is 0 
   the result is placed in the W register. If 'd' is
   1 the result is stored back in register 'f'.
   
   STATUS: C
   CYCLES: 1
   
   0 <= f <= 127
   d [0,1]
   
    -->[C]-->[REGISTER F]-->
   |                        |
    ------------------------
*/
void 
Pic::RRF( unsigned opcode, int operand )
{
  PicRegister *cell = m_ram->getRegister( operand );
  int value = cell->value();
  
  const bool tmpCarry = (value & 0x01);
  value >>= 1;
  
  if (carryFlag())
  {
    value |= 0x80;
  }
  
  setStatusFlag( C, tmpCarry );
   
  storeResult( value, cell, opcode );
  
  traceInstruction( cell->valuePic() );
}

/* SUBWF f,d - Subtract W from f

   Subtract (2's complement method) W register
   from register 'f'. If 'd' is 0 the result is
   stored in the W register. If 'd' is 1 the
   result is stored back in register 'f'.
   
   STATUS: C,DC,Z
   CYCLES: 1
   
   0 <= f <= 127
   d [0,1]
   
   (f)-(W) -> (dest)
   
   TODO: DC
*/
void  
Pic::SUBWF( unsigned opcode, int operand )
{
  PicRegister *cell = m_ram->getRegister( operand );
  const int value = cell->value() - m_w;
  
  setC_Z_DC_Sub( cell->valuePic(), value, m_w );  
  
  storeResult( value, cell, opcode );
  
  traceInstruction( cell->valuePic() );
}

/* SWAPF f,d - Swap Nibbles in f

   The upper and lower nibbles of register 'f'
   are exchanged. If 'd' is 0 the result is placed
   in W register. If 'd' is 1 the result is placed
   in register 'f'.
   
   STATUS: none
   CYCLES: 1
   
   0 <= f <= 127
   d [0,1]
   
   (f<3:0>) -> (dest<7:4>),
   (f<7:4>) -> (dest<3:0>)
*/
void 
Pic::SWAPF( unsigned opcode, int operand )
{
  PicRegister *cell = m_ram->getRegister( operand );
  int value = cell->value();  
  value = ((value >> 4) & 0x0f) | ((value << 4) & 0xf0);
  
  storeResult( value, cell, opcode );
  
  traceInstruction( cell->valuePic() );
}

/* XORWF f,d - Exclusive OR W with f

   Exclusive OR the contents of the W register
   'f'. If 'd' is 0 the result is stored in the W
   register. If 'd' is 1 the result is stored back
   in register 'f'
   
   STATUS: Z
   CYCLES: 1
   
   0 <= f <= 127
   d [0,1]
   
   (W) .XOR. (f) -> (dest)
*/
void 
Pic::XORWF( unsigned opcode, int operand )
{
  PicRegister *cell = m_ram->getRegister( operand );
  const int value = m_w ^ cell->value();
    
  setZ( value );
  
  storeResult( value, cell, opcode );
  
  traceInstruction( cell->valuePic() );
}
  
///////////////////////////////////////////////
//bit oriented file register operation
///////////////////////////////////////////////

/* BCF f,b - Bit Set f

   Bit 'b' in register 'f' is cleared.
   
   STATUS: none
   CYCLES: 1
   
   0 <= f <= 127
   0 <= b <= 7
   
   0 -> (f<b>)
*/
void 
Pic::BCF( unsigned opcode, int operand )
{  
  PicRegister *cell = m_ram->getRegister( operand );
  cell->clearBit( bit( opcode ) );
  
  traceInstruction( cell->valuePic() );
}

/* BSF f,b - Bit Set f

   Bit 'b' in register 'f' is set.
   
   STATUS: none
   CYCLES: 1
   
   0 <= f <= 127
   0 <= b <= 7
   
   1 -> (f<b>)
*/
void 
Pic::BSF( unsigned opcode, int operand )
{
  PicRegister *cell = m_ram->getRegister( operand );
  // user write
  cell->setBit( bit( opcode ) );
  
  traceInstruction( cell->valuePic() );
}

/* BTFSC f,b - Bit Test f, Skip if Clear

   If bit 'b' in register 'f' is '0' then the
   next instruction is skipped. If bit 'b' is
   '0' then the next instruction fetched during
   the current instruction execution is discarded,
   and a NOP is executed instead, making this
   a two cycle instruction
   
   STATUS: none
   CYCLES: 1(2)
   
   0 <= f <= 127
   0 <= b <= 7
   
   skip if (f<b>) = 0
*/
void 
Pic::BTFSC( unsigned opcode, int operand )
{
  PicRegister *cell = m_ram->getRegister( operand );
  //const int pcAtCall = m_pc;
  //const unsigned cyclesAtCall = m_cycles;
  
  // user read
  if (!(cell->bitSet( bit( opcode ) )))
  {
    // this skips the next instruction
    //
    flushInstructionPipeline();   
  }
  
  traceInstruction( cell->valuePic() );
}

/* BTFSS f,b - Bit Test f, Skip if Set

   If bit 'b' in register 'f' is '1' then the next
   instruction is skipped. If bit 'b' is '1'
   then the next instruction fetched during
   the current instruction execution is discarded,
   and a NOP is executed instead, making this
   a two cycle instruction.
   
   STATUS: none
   CYCLES 1(2)
   
   0 <= f <= 127
   0 <= b <= 7
   
   skip if (f<b>) = 1
*/
void 
Pic::BTFSS( unsigned opcode, int operand )
{
  PicRegister *cell = m_ram->getRegister( operand );
  //const int pcAtCall = m_pc;
  //const unsigned cyclesAtCall = m_cycles;
  
  // user read
  if (cell->bitSet( bit( opcode ) ))
  {
    // this skips the next instruction
    //
    flushInstructionPipeline();
  }
  
  traceInstruction( cell->valuePic() );
}

// literal and control operations
//
/* ADDLW k - Add Literal and W

   The contents of the W register are added to the
   eight bit literal 'k' and the result is placed
   in the W register
   
   STATUS: C, DC, Z
   CYCLES: 1
   
   0 <= k <= 255
   
   (W) + k -> (W)
   
   TODO: DC
*/
void 
Pic::ADDLW( unsigned, int operand )
{
  const int oldValue = m_w;
  
  m_w += operand;
  
  setC_Z_DC( oldValue, m_w, operand );
  
  m_w &= 0xff;
  
  traceInstruction();
}

/* ANDLW k - AND Literal and W

   The contents of W are AND'ed with the eight bit
   literal 'k'. The result is placed in the W register
   
   STATUS: Z
   CYCLES: 1
   
   0 <= k <= 255
   
   (W) .AND. (k) -> (W)
*/
void 
Pic::ANDLW( unsigned, int operand )
{
  m_w &= operand;
  
  setZ( m_w );
  
  traceInstruction();
}

/* CALL k - Call Subroutine

   Call Subroutine. First, return address (PC+1)
   is pushed onto the stack. The eleven bit
   immediate address is loaded into PC bits <10:0>.
   The upper bits of the PC are loaded form PCLATH.
   CALL is a two-cycle instruction.
   
   STATUS: none
   CYCLES: 2
   
   0 <= k <= 2047
   
   (PC)+1 -> TOS,
   k -> PC<10:0>,
   (PCLATH<4:3>) -> PC<12:11>
 */
void 
Pic::CALL( unsigned, int operand )
{
  m_callDepth++;
  
  // call statistics
  //
  m_stack->setCyclesAtCall( m_cycles );
  m_stack->setPcAtCall( m_tracePc ); 
  m_stack->setCall( operand );
  
  // call itself
  //
  m_stack->push( m_pc );
 
  const int pclath = (m_sfr[PCLATH]->valuePic() & 0x18) << 8;
  m_pc = operand | pclath;

  flushInstructionPipeline();
  
  // fetch next instruction (or we skip the instruction at PC)
  //
  fetchInstruction( m_pc );  
  
  traceInstruction();
}

/* GOTO k - Unconditional Branch

   GOTO is an unconditional branch. The eleven
   bit immediate value is loaded into PC bits <10:0>.
   The upper bits of PC are loaded from PCLATH<4:3>.
   GOTO is a two-cycle instruction.
   
   STATUS: none
   CYCLES: 2
   
   0 <= k <= 2047
   
   k -> PC<10:0>
   PCLATH<4:3> -> PC<12:11>
 */
void 
Pic::GOTO( unsigned, int operand )
{
  m_pc = operand;
  const int pclath = (m_sfr[PCLATH]->valuePic() & 0x18) << 8;
  m_pc |= pclath;
  
  flushInstructionPipeline();

  // fetch next instruction (or we skip the instruction at PC)
  //
  fetchInstruction( m_pc );  
  
  traceInstruction();
}

/* IORLW k - Inclusive OR Literal with W

   The contents of the W register is OR'ed with the
   eight bit literal 'k'. The result is placed in
   the W register.
   
   STATUS: Z
   CYCLES: 1
   
   0 <= k <= 255
   
   (W) .OR. k -> (W)
*/
void 
Pic::IORLW( unsigned, int operand )
{
  m_w |= operand;
  
  setZ( m_w );
  
  traceInstruction();
}

/* MOVLW k - Move Literal to W

   The eight bit literal 'k' is loaded into W register.

   STATUS: none
   CYCLES: 1
   
   0 <= k <= 255
   
   k -> (W)
*/
void 
Pic::MOVLW( unsigned, int operand )
{
  m_w = operand;
  
  traceInstruction();
}

/* RETFIE - Return from Interrupt

   Return from Interrupt. Stack is POPed and Top of
   Stack (TOS) is loaded in the PC. Interrupts are
   enabled by setting Global Interrupt Enable bit, 
   GIE (INTCON<7>). This is a two-cycle instruction.
   
   STATUS: none
   CYCLES: 2
   
   TOS -> PC
   1 -> GIE
*/
void 
Pic::RETFIE( unsigned, int  )
{  
  m_pc = m_stack->pop();
  
  m_sfr[INTCON]->setBitPic( 7 );
  
  flushInstructionPipeline();

  // fetch next instruction (or we skip the instruction at PC)
  //
  fetchInstruction( m_pc );  
  
  traceInstruction();
}

/* CLRWDT - Clear Watchdog Timer

   CLRWDT instruction resets the Watchdog Timer.
   It also resets the prescaler of the WDT.
   STATUS bits /TO and /PD are set.
   
   STATUS: /TO,/PD
   CYCLES: 1
   
   00h -> WDT
   0 -> WDT prescaler,
   1 -> /TO
   1 -> /PD
   
   TODO: WDT
*/
void 
Pic::CLRWDT( unsigned, int  )
{
  m_sfr[STATUS]->setBitPic( 3 );
  m_sfr[STATUS]->setBitPic( 4 );
  
  traceInstruction();
}

/* RETLW k - Return with Literal in W

   The W register is loaded with the eight literal 'k'.
   The program counter is loaded from the top of the
   stack (the return address). This is a two-cycle
   instruction.
   
   STATUS: none
   CYCLES: 2
   
   0 <= k <= 255
   
   k -> (W)
   TOS -> PC
*/
void 
Pic::RETLW( unsigned, int operand )
{  
  if (m_callDepth>0) m_callDepth--;
  
  m_w = operand;  
  m_pc = m_stack->pop();
  
  flushInstructionPipeline();
  
  // fetch next instruction (or we skip the instruction at PC)
  //
  fetchInstruction( m_pc );  
  
  callStatistics( m_stack->pcAtCall(), m_cycles-m_stack->cyclesAtCall() );
  
  traceInstruction();
}

/* RETURN - Return from Subroutine

   Return from subroutine. The stack is POPed and the top
   of the stack (TOS) is loaded into the program counter. 
   This is a two-cycle instruction.
   
   STATUS: none
   CYCLES: 2
   
   TOS -> PC
*/
void 
Pic::RETURN( unsigned, int  )
{
  if (m_callDepth>0) m_callDepth--;
  
  m_pc = m_stack->pop();
  
  flushInstructionPipeline();
  
  // fetch next instruction (or we skip the instruction at PC)
  //
  fetchInstruction( m_pc );  
  
  callStatistics( m_stack->pcAtCall(), m_cycles-m_stack->cyclesAtCall() );
  
  traceInstruction();
}

/* SLEEP

   The power-down STATUS bit /PD is cleared. Timeout
   STATUS bit /TO is set. Watchdog timer and its
   prescaler are cleared. The processor is put into
   SLEEP mode with the oscillator stopped.
   
   STATUS: /TO,/PD
   CYCLES: 1
   
   00h -> WDT
   0 -> WDT prescaler
   1 -> /TO
   0 -> /PD
   
   TODO: WDT
*/
void 
Pic::SLEEP( unsigned, int  )
{
  m_sleep = true;
    
  m_sfr[STATUS]->setBitPic( 4 );
  m_sfr[STATUS]->clearBitPic( 3 );
  
  traceInstruction();
}

/* SUBLW k - Subtract W from Literal

   The W register is subtracted (2's complement method)
   from the eight bit literal 'k'. The result is placed
   in the W register.
   
   STATUS: C,DC,Z
   CYCLES: 1
   
   0 <= k <= 255
   
   k-(W) -> (W)
   
   TODO: DC
*/
void 
Pic::SUBLW( unsigned, int operand )
{
  const int oldValue = m_w;
  
  m_w = operand - m_w;
  
  setC_Z_DC_Sub( oldValue, m_w, operand );
  
  m_w &= 0xff;
  
  traceInstruction();
}

/* XORLW k - Exclusive OR Literal with W

   The contents of the W register are XORed with the
   eight bit literal 'k'. The result is placed in the
   W register.
   
   STATUS: Z
   CYCLES: 1
   
   0 <= k <= 255
   
   (W) .XOR. k-> (W)
*/
void 
Pic::XORLW( unsigned, int operand )
{
  m_w ^= operand;
  m_w &= 0xff;

  setZ( m_w );
    
  traceInstruction();
}

void 
Pic::OPTION( unsigned , int  )
{
  m_sfr[OPTION_REG]->setValue( m_w );
    
  traceInstruction();
}

void 
Pic::TRIS( unsigned opcode, int  )
{
  int reg = opcode & 0x07;
  
  switch (reg)
  {
  case 5:
    m_sfr[TRISA]->setValue( m_w );
    break;
  case 6:
    m_sfr[TRISB]->setValue( m_w );
    break;
  }
  
  traceInstruction();
}

void
Pic::compile()
{
  for (int i=0; i<m_rom->dataCnt(); ++i)
  {
    CodeTable *code = findCodeFromOpcode( m_rom->at(i) );
    
    if (code)
    {
/*      QString addr;
      addr.sprintf( "%03X ", i );
      std::cerr << addr.latin1()
                << "compile " << opcodeToString( m_rom->at( i ) ).latin1()
                << " -> " << code->mnemonic.latin1();
      */
      m_compiledFunction[i] = code->function;

      switch (code->mode)
      {
      case File:
        m_compiledOperand[i] = file( m_rom->at(i) );
        //std::cerr << " FILE";
        break;
      case Address:
        m_compiledOperand[i] = address( m_rom->at(i) );
        //std::cerr << " ADDRESS";
        break;
      case Literal:
        m_compiledOperand[i] = literal( m_rom->at(i) );
        //std::cerr << " LITERAL";
        break;
      default:
        m_compiledOperand[i] = 0;
        //std::cerr << " Don't care";
        break;
      }
      
      //std::cerr << std::endl;
    }
    else
    {
      std::cerr << "Complie fuckup at address " << i << std::endl;
      ::exit(-1);
    }
  }
}

CodeTable *
Pic::findCodeFromOpcode( unsigned opcode )
{
  for (int o=0; o<m_numOpcode; ++o)
  {
    CodeTable *code = &m_codeTable[o];

    if ((opcode & code->mask) == code->opcode)
    {
      return code;
    }
  }
  
  return 0;
}

QString
Pic::opcodeToString( unsigned opcode )
{
  return DisAss::toString( opcode );
}

bool
Pic::isActiveBreakpoint() const
{
  return m_rom->isActiveBreakpoint( m_pc );
}

void
Pic::createCodeTable()
{
  m_codeTable = new CodeTable [m_numOpcode];
  
  // byte oriented file register operations
  //
  m_codeTable[0].mnemonic = "addwf";
  m_codeTable[0].opcode   = 0x0700;
  m_codeTable[0].mask     = 0x3f00;
  m_codeTable[0].function = &Pic::ADDWF;
  m_codeTable[0].mode     = Pic::File;
  
  m_codeTable[1].mnemonic = "andwf";
  m_codeTable[1].opcode   = 0x0500;
  m_codeTable[1].mask     = 0x3f00;
  m_codeTable[1].function = &Pic::ANDWF;
  m_codeTable[1].mode     = Pic::File;
  
  m_codeTable[2].mnemonic = "clrf";
  m_codeTable[2].opcode   = 0x0180;
  m_codeTable[2].mask     = 0x3f80;
  m_codeTable[2].function = &Pic::CLRF;
  m_codeTable[2].mode     = Pic::File;
  
  m_codeTable[3].mnemonic = "clrw";
  m_codeTable[3].opcode   = 0x0100;
  m_codeTable[3].mask     = 0x3f00;
  m_codeTable[3].function = &Pic::CLRW;
  m_codeTable[3].mode     = Pic::DontCare;
  
  m_codeTable[4].mnemonic = "comf";
  m_codeTable[4].opcode   = 0x0900;
  m_codeTable[4].mask     = 0x3f00;
  m_codeTable[4].function = &Pic::COMF;
  m_codeTable[4].mode     = Pic::File;
  
  m_codeTable[5].mnemonic = "decf";
  m_codeTable[5].opcode   = 0x0300;
  m_codeTable[5].mask     = 0x3f00;
  m_codeTable[5].function = &Pic::DECF;
  m_codeTable[5].mode     = Pic::File;
  
  m_codeTable[6].mnemonic = "decfsz";
  m_codeTable[6].opcode   = 0x0b00;
  m_codeTable[6].mask     = 0x3f00;
  m_codeTable[6].function = &Pic::DECFSZ;
  m_codeTable[6].mode     = Pic::File;
  
  m_codeTable[7].mnemonic = "incf";
  m_codeTable[7].opcode   = 0x0a00;
  m_codeTable[7].mask     = 0x3f00;
  m_codeTable[7].function = &Pic::INCF;
  m_codeTable[7].mode     = Pic::File;
  
  m_codeTable[8].mnemonic = "incfsz";
  m_codeTable[8].opcode   = 0x0f00;
  m_codeTable[8].mask     = 0x3f00;
  m_codeTable[8].function = &Pic::INCFSZ;
  m_codeTable[8].mode     = Pic::File;
  
  m_codeTable[9].mnemonic = "iorwf";
  m_codeTable[9].opcode   = 0x0400;
  m_codeTable[9].mask     = 0x3f00;
  m_codeTable[9].function = &Pic::IORWF;
  m_codeTable[9].mode     = Pic::File;
  
  m_codeTable[10].mnemonic = "movf";
  m_codeTable[10].opcode   = 0x0800;
  m_codeTable[10].mask     = 0x3f00;
  m_codeTable[10].function = &Pic::MOVF;
  m_codeTable[10].mode     = Pic::File;
  
  m_codeTable[11].mnemonic = "movwf";
  m_codeTable[11].opcode   = 0x0080;
  m_codeTable[11].mask     = 0x3f80;
  m_codeTable[11].function = &Pic::MOVWF;
  m_codeTable[11].mode     = Pic::File;
  
  m_codeTable[12].mnemonic = "nop";
  m_codeTable[12].opcode   = 0x0000;
  m_codeTable[12].mask     = 0x3f9f;
  m_codeTable[12].function = &Pic::NOP;
  m_codeTable[12].mode     = Pic::DontCare;
  
  m_codeTable[13].mnemonic = "rlf";
  m_codeTable[13].opcode   = 0x0d00;
  m_codeTable[13].mask     = 0x3f00;
  m_codeTable[13].function = &Pic::RLF;
  m_codeTable[13].mode     = Pic::File;
  
  m_codeTable[14].mnemonic = "rrf";
  m_codeTable[14].opcode   = 0x0c00;
  m_codeTable[14].mask     = 0x3f00;
  m_codeTable[14].function = &Pic::RRF;
  m_codeTable[14].mode     = Pic::File;
  
  m_codeTable[15].mnemonic = "subwf";
  m_codeTable[15].opcode   = 0x0200;
  m_codeTable[15].mask     = 0x3f00;
  m_codeTable[15].function = &Pic::SUBWF;
  m_codeTable[15].mode     = Pic::File;
  
  m_codeTable[16].mnemonic = "swapf";
  m_codeTable[16].opcode   = 0x0e00;
  m_codeTable[16].mask     = 0x3f00;
  m_codeTable[16].function = &Pic::SWAPF;
  m_codeTable[16].mode     = Pic::File;
  
  m_codeTable[17].mnemonic = "xorwf";
  m_codeTable[17].opcode   = 0x0600;
  m_codeTable[17].mask     = 0x3f00;
  m_codeTable[17].function = &Pic::XORWF;
  m_codeTable[17].mode     = Pic::File;
  
  // bit oriented file register operations
  //
  m_codeTable[18].mnemonic = "bcf";
  m_codeTable[18].opcode   = 0x1000;
  m_codeTable[18].mask     = 0x3c00;
  m_codeTable[18].function = &Pic::BCF;
  m_codeTable[18].mode     = Pic::File;
  
  m_codeTable[19].mnemonic = "bsf";
  m_codeTable[19].opcode   = 0x1400;
  m_codeTable[19].mask     = 0x3c00;
  m_codeTable[19].function = &Pic::BSF;
  m_codeTable[19].mode     = Pic::File;
  
  m_codeTable[20].mnemonic = "btfsc";
  m_codeTable[20].opcode   = 0x1800;
  m_codeTable[20].mask     = 0x3c00;
  m_codeTable[20].function = &Pic::BTFSC;
  m_codeTable[20].mode     = Pic::File;
  
  m_codeTable[21].mnemonic = "btfss";
  m_codeTable[21].opcode   = 0x1c00;
  m_codeTable[21].mask     = 0x3c00;
  m_codeTable[21].function = &Pic::BTFSS;
  m_codeTable[21].mode     = Pic::File;
  
  // literal and control operations
  //
  m_codeTable[22].mnemonic = "addlw";
  m_codeTable[22].opcode   = 0x3e00;
  m_codeTable[22].mask     = 0x3e00;
  m_codeTable[22].function = &Pic::ADDLW;
  m_codeTable[22].mode     = Pic::Literal;
  
  m_codeTable[23].mnemonic = "andlw";
  m_codeTable[23].opcode   = 0x3900;
  m_codeTable[23].mask     = 0x3f00;
  m_codeTable[23].function = &Pic::ANDLW;
  m_codeTable[23].mode     = Pic::Literal;
  
  m_codeTable[24].mnemonic = "call";
  m_codeTable[24].opcode   = 0x2000;
  m_codeTable[24].mask     = 0x3800;
  m_codeTable[24].function = &Pic::CALL;
  m_codeTable[24].mode     = Pic::Address;
  
  m_codeTable[25].mnemonic = "clrwdt";
  m_codeTable[25].opcode   = 0x0064;
  m_codeTable[25].mask     = 0x3fff;
  m_codeTable[25].function = &Pic::CLRWDT;
  m_codeTable[25].mode     = Pic::DontCare;
  
  m_codeTable[26].mnemonic = "goto";
  m_codeTable[26].opcode   = 0x2800;
  m_codeTable[26].mask     = 0x3800;
  m_codeTable[26].function = &Pic::GOTO;
  m_codeTable[26].mode     = Pic::Address;
  
  m_codeTable[27].mnemonic = "iorlw";
  m_codeTable[27].opcode   = 0x3800;
  m_codeTable[27].mask     = 0x3f00;
  m_codeTable[27].function = &Pic::IORLW;
  m_codeTable[27].mode     = Pic::Literal;
  
  m_codeTable[28].mnemonic = "movlw";
  m_codeTable[28].opcode   = 0x3000;
  m_codeTable[28].mask     = 0x3c00;
  m_codeTable[28].function = &Pic::MOVLW;
  m_codeTable[28].mode     = Pic::Literal;
  
  m_codeTable[29].mnemonic = "retfie";
  m_codeTable[29].opcode   = 0x0009;
  m_codeTable[29].mask     = 0x3fff;
  m_codeTable[29].function = &Pic::RETFIE;
  m_codeTable[29].mode     = Pic::DontCare;
  
  m_codeTable[30].mnemonic = "retlw";
  m_codeTable[30].opcode   = 0x3400;
  m_codeTable[30].mask     = 0x3f00;
  m_codeTable[30].function = &Pic::RETLW;
  m_codeTable[30].mode     = Pic::Literal;
  
  m_codeTable[31].mnemonic = "return";
  m_codeTable[31].opcode   = 0x0008;
  m_codeTable[31].mask     = 0x3fff;
  m_codeTable[31].function = &Pic::RETURN;
  m_codeTable[31].mode     = Pic::DontCare;
  
  m_codeTable[32].mnemonic = "sleep";
  m_codeTable[32].opcode   = 0x0063;
  m_codeTable[32].mask     = 0x3fff;
  m_codeTable[32].function = &Pic::SLEEP;
  m_codeTable[32].mode     = Pic::DontCare;
  
  m_codeTable[33].mnemonic = "sublw";
  m_codeTable[33].opcode   = 0x3c00;
  m_codeTable[33].mask     = 0x3f00;
  m_codeTable[33].function = &Pic::SUBLW;
  m_codeTable[33].mode     = Pic::Literal;
  
  m_codeTable[34].mnemonic = "xorlw";
  m_codeTable[34].opcode   = 0x3a00;
  m_codeTable[34].mask     = 0x3f00;
  m_codeTable[34].function = &Pic::XORLW;
  m_codeTable[34].mode     = Pic::Literal;
  
  m_codeTable[35].mnemonic = "option";
  m_codeTable[35].opcode   = 0x0062;
  m_codeTable[35].mask     = 0x3fff;
  m_codeTable[35].function = &Pic::OPTION;
  m_codeTable[35].mode     = Pic::DontCare;
  
  m_codeTable[36].mnemonic = "tris";
  m_codeTable[36].opcode   = 0x0060;
  m_codeTable[36].mask     = 0x3ff8;
  m_codeTable[36].function = &Pic::TRIS;
  m_codeTable[36].mode     = Pic::DontCare;
}

void
Pic::createRAM()
{  
  m_ram = new PicRAM( 4, this, &m_pc );
  
  // get the hardcoded special register
  //
  m_sfr[INDF] = m_ram->indf();
  m_sfr[FSR] = m_ram->fsr();
  m_sfr[STATUS] = m_ram->status();
  m_sfr[PCL] = m_ram->pcl();
  m_sfr[OPTION_REG] = m_ram->option();
  
  // 80 register in bank 0 and 1
  //
  m_ram->createGeneralPurpose( 0, 32, 80 );
  m_ram->createGeneralPurpose( 1, 32, 80 );
  
  // 48 register in bank 2
  //
  m_ram->createGeneralPurpose( 2, 32, 48 );
  
  // 16 register with shadow in all banks
  //
  m_ram->createGeneralPurpose( 0, 112, 16 );
  m_ram->createShadow( 1, 112, 16, 0 );  
  m_ram->createShadow( 2, 112, 16, 0 );  
  m_ram->createShadow( 3, 112, 16, 0 );  
  
  m_sfr[TMR0] = m_ram->createRegister( 0, 1, PicRegister::SFR, "TMR0" ); 
  m_ram->createShadow( 2, 1, 1, 0 ); 

  m_sfr[PORTA] = m_ram->createRegister( 0, 5, PicRegister::SFR, "PORTA" ); 
  m_sfr[PORTA]->setFlagNames( "RA0", "RA1", "RA2", "RA3", "RA4", 
                              "RA5", "RA6", "RA7" );
  
  m_sfr[PORTB] = m_ram->createRegister( 0, 6, PicRegister::SFR, "PORTB" ); 
  m_sfr[PORTB]->setFlagNames( "RB0", "RB1", "RB2", "RB3", "RB4", 
                              "RB5", "RB6", "RB7" );
  m_ram->createShadow( 2, 6, 1, 0 ); 

  m_sfr[TRISA] = m_ram->createRegister( 1, 5, PicRegister::SFR, "TRISA",
                                        0xff, 0xff ); 
  
  m_sfr[TRISB] = m_ram->createRegister( 1, 6, PicRegister::SFR, "TRISB",
                                        0xff, 0xff ); 
  m_ram->createShadow( 3, 6, 1, 1 ); 
  
  m_sfr[PCLATH] = m_ram->createRegister( 0, 10, PicRegister::SFR, 
                                         "PCLATH", 0x1f ); 
  m_ram->createShadow( 1, 10, 1, 0 ); 
  m_ram->createShadow( 2, 10, 1, 0 ); 
  m_ram->createShadow( 3, 10, 1, 0 ); 
  
  m_sfr[INTCON] = m_ram->createRegister( 0, 11, PicRegister::SFR, "INTCON" ); 
  m_sfr[INTCON]->setFlagNames( "RBIF", "INTF", "T0IF", "RBIE", "INTE", 
                               "T0IE", "PEIE", "GIE" );
  m_ram->createShadow( 1, 11, 1, 0 ); 
  m_ram->createShadow( 2, 11, 1, 0 ); 
  m_ram->createShadow( 3, 11, 1, 0 ); 
  
  m_sfr[PIR1] = m_ram->createRegister( 0, 12, PicRegister::SFR, "PIR1", 0xf7 ); 
  m_sfr[PIR1]->setFlagNames( "TMR1IF", "TMR2IF", "CCP1IF", "n/a", 
                             "TXIF", "RCIF", "CMIF", "EEIF" );
  
  m_sfr[TMR1L] = m_ram->createRegister( 0, 14, PicRegister::SFR, "TMR1L" ); 
  m_sfr[TMR1H] = m_ram->createRegister( 0, 15, PicRegister::SFR, "TMR1H" ); 
  
  m_sfr[T1CON] = m_ram->createRegister( 0, 0x10, PicRegister::SFR, 
                                        "T1CON", 0x3f ); 
  m_sfr[T1CON]->setFlagNames( "TMR1ON", "TMR1CS", "/T1SYNC", "T1OSCEN", 
                              "T1CKPS0", "T1CKPS1", "n/a", "n/a" );
  
  m_sfr[TMR2] = m_ram->createRegister( 0, 0x11, PicRegister::SFR, 
                                        "TMR2" ); 
  
  m_sfr[T2CON] = m_ram->createRegister( 0, 0x12, PicRegister::SFR, 
                                        "T2CON", 0x7f ); 
  m_sfr[T2CON]->setFlagNames( "T2CKPS0", "T2CKPS1", "TMR2ON", "TOUTPS0", 
                              "TOUTPS1", "TOUTPS2", "TOUTPS3", "n/a" );

  m_sfr[CCPR1L] = m_ram->createRegister( 0, 0x15, PicRegister::SFR, 
                                        "CCPR1L" ); 
  m_sfr[CCPR1H] = m_ram->createRegister( 0, 0x16, PicRegister::SFR, 
                                        "CCPR1H" ); 
  
  m_sfr[CCP1CON] = m_ram->createRegister( 0, 0x17, PicRegister::SFR, 
                                        "CCP1CON", 0x3f ); 
  m_sfr[CCP1CON]->setFlagNames( "CCP1M0", "CCP1M1", "CCP1M2", "CCP1M3", 
                                "CCP1Y", "CCP1X", "n/a", "n/a" );
  
  m_sfr[RCSTA] = m_ram->createRegister( 0, 0x18, PicRegister::SFR, 
                                        "RCSTA" ); 
  m_sfr[RCSTA]->setFlagNames( "RX9D", "OERR", "FERR", "ADEN", 
                              "CREN", "SREN", "RX9", "SPEN" );
  
  m_sfr[TXREG] = m_ram->createRegister( 0, 0x19, PicRegister::SFR, 
                                        "TXREG" ); 
  
  m_sfr[RCREG] = m_ram->createRegister( 0, 0x1a, PicRegister::SFR, 
                                        "RCREG" ); 
  
  m_sfr[CMCON] = m_ram->createRegister( 0, 0x1f, PicRegister::SFR, 
                                        "CMCON" ); 
  m_sfr[CMCON]->setFlagNames( "CM0", "CM1", "CM2", "CIS", 
                              "C1INV", "C2INV", "C1OUT", "C2OUT" );
  
  m_sfr[PIE1] = m_ram->createRegister( 1, 0x0c, PicRegister::SFR, 
                                        "PIE1", 0xf7 ); 
  m_sfr[PIE1]->setFlagNames( "TMR1IE", "TMR2IE", "CCP1IE", "n/a", 
                             "TXIE", "RCIE", "CMIE", "EEIE" );
 
  m_sfr[PCON] = m_ram->createRegister( 1, 0x0e, PicRegister::SFR, 
                                        "PCON", 0x0b, 0x08 ); 
  m_sfr[PCON]->setFlagNames( "/BOD", "/POR", "n/a", "CIS", 
                              "n/a", "n/a", "n/a", "n/a" );

 
  m_sfr[PR2] = m_ram->createRegister( 1, 0x12, PicRegister::SFR, 
                                        "PR2", 0xff, 0xff ); 
 
  m_sfr[TXSTA] = m_ram->createRegister( 1, 0x18, PicRegister::SFR, 
                                        "TXSTA", 0xf7, 0x02 ); 
  m_sfr[TXSTA]->setFlagNames( "TX9D", "TRMT", "BRGH", "n/a", 
                              "SYNC", "TXEN", "TX9", "CSRC" );

}

void
Pic::createTimer()
{
  m_numTimer = 2;
  
  m_timer[0] = new PicTimer0( this );
  
  PicTimer16 *t2 = new PicTimer16( "TIMER1", m_sfr[T1CON], 0 );
  m_timer[1] = t2;
  t2->setRegister( m_sfr[TMR1L], m_sfr[TMR1H] );
  
/*  PicTimer8 *t3 = new PicTimer8( "TIMER2" );
  m_timer[2] = t3;
  t3->setRegister( m_sfr[TMR2] );*/
}
