//======================================================================
// File:		pic.h
// Author:	Matthias Toussaint
// Created:	Mon Jul 26 23:08:00 CET 2004
//----------------------------------------------------------------------
// Permission to use, copy, modify, and distribute this software and its
// documentation  for any  purpose and  without fee is  hereby  granted,
// provided  that below copyright notice appear  in all copies  and that
// both  that  copyright  notice and  this permission  notice  appear in
// supporting documentation.
// 
// This  file is  provided AS IS  with no  warranties  of any kind.  The
// author shall  have no liability  with respect  to the infringement of
// copyrights, trade  secrets  or any patents by  this file  or any part
// thereof.  In no event will the author be liable  for any lost revenue
// or profits or other special, indirect and consequential damages.
//----------------------------------------------------------------------
// (c) 2004 Matthias Toussaint
//======================================================================

#ifndef PIC_HH
#define PIC_HH

#include <qdatetime.h>
#include <qsemaphore.h>
#include <qstring.h>
#include <qthread.h>
#include <qvaluelist.h>

#include <disass.h>
#include <picram.h>
#include <picregister.h>
#include <picrom.h>
#include <picstack.h>
#include <symboltable.h>
#include <instructiontracebuffer.h>
#include <registertrace.h>

#include <iostream>

class Pic;
class PicTimer;

struct CodeTable
{
  QString   mnemonic;
  unsigned  opcode;
  unsigned  mask;
  void (Pic::*function)( unsigned, int );
  int       mode;
  
};

class Pic : public QThread
{
public:
  enum OperandMode
  {
    File,
    Address,
    Literal,
    DontCare
  };
  
  enum StautsFlags
  {
    C=0,
    DC,
    Z,
    nPD,
    nTO,
    RP0,
    RP1,
    IRP
  };
    
  enum StatusBits
  {
    C_BIT   = 0x01,
    DC_BIT  = 0x02,
    Z_BIT   = 0x04,
    nPD_BIT = 0x08,
    nTO_BIT = 0x10,
    RP0_BIT = 0x20,
    RP1_BIT = 0x40,
    IRP_BIT = 0x80
  };
    
  enum Register
  {
    ADCOND0=0,
    ADCON1,
    INDF,
    ADRES,
    CCP1CON,
    CCP2CON,
    CCPR1H,
    CCPR1L,
    CCPR2H,
    CCPR2L,
    CMCON,
    FSR,
    INTCON,
    OPTION_REG,
    PCL,
    PCLATH,
    PCON,
    PIE1,
    PIR1,
    PORTA,
    PORTB,
    PR2,
    RCREG,
    RCSTA,
    SSPBUF,
    SSPCON,
    SSPADD,
    SSPATAT,
    SPBRG,
    STATUS,
    T1CON,
    T2CON,
    TMR0,
    TMR1H,
    TMR1L,
    TMR2,
    TRISA,
    TRISB,
    TXREG,
    TXSTA,
    OSCAL,
    NUM_SFR
  };
     
  enum TIMER
  {
    TIMER0=0,
    TIMER1,
    TIMER2,
    NumTimer
  };
    
  enum Operand
  {
    NO_OPERAND  = 0x00,
    FILE        = 0x01,
    DESTINATION = 0x02,
    BIT         = 0x04,
    ADDRESS     = 0x08,
    LITERAL     = 0x10,
  };

  Pic();
  ~Pic();
    
  void reset(); 
  void compile();

  void run();
  
  /// Convenience function for statusbit retrieval
  bool statusBit( int bit ) const { return m_status->bitSetPic( bit ); }
  /// Convenience function for zeroflag
  bool zeroFlag() const { return (m_status->valuePic() & Z_BIT); }
  /// Convenience function for carry flag
  bool carryFlag() const { return (m_status->valuePic() & C_BIT); }
  /// Convenience function for decimal carryflag
  bool dcFlag() const { return (m_status->valuePic() & DC_BIT); }
  /// Convenience function to set status bit
  void setStatusFlag( int, bool );
  
  bool sleep() const { return m_sleep; }
  
  /// Halt Simulation
  void halt();
  /// Returns true if halted
  bool halted() const;
  
  /// Execute one step
  void step();
  /// Continue simulation (e.g. run)
  void cont();
  /// Step over call
  void over();
  /// Finish subroutine
  void finishCall();
  /// Terminates thread
  void quit();
  
  void lock();
  void unlock();
    
  /// Returns number of cycles since last counter reset
  unsigned cycles() const { return m_cycles; }
  /// Clear cycle counter
  void clearCycles();
  
  /// Return content of working register
  int workingRegister() const { return m_w; }
  /// Set PC from outside
  void setPc( int pc, bool setPcl=true );
  /// Return current pc
  int pc() const { return m_pc; }
  /// Retrieve pc from PCL register
  int pcFromPCL() const;
  
  /// Load ROM with given data
  void loadROM( unsigned *, int );
  
  /// Return pointer to RAM object
  PicRAM   *ram()   const { return m_ram; }
  /// Return pointer to ROM object
  PicROM   *rom()   const { return m_rom; }
  /// Return pointer to stack object
  PicStack *stack() const { return m_stack; }
  /// Return pointer to symbol table
  SymbolTable *symbolTable() const { return m_symbolTable; }
  /// Return pointer to instruction trace buffer
  InstructionTraceBuffer *instructionTraceBuffer() const { return m_instructionTraceBuffer; }
  RegisterTrace *registerTrace() const { return m_registerTrace; }
  
  PicRegister *sfr( int i ) { return m_sfr[i]; }
  
  QString opcodeToString( unsigned opcode );
  
  /// Compute and return current clock speed
  double clockSpeed();
  /// Return crystal frequency
  double crystalFrequency() const { return m_crystalFrequency; }
  
  void setCrystalFrequency( double f );
  /// Return speed in percent relative to clock frequency
  double percentSpeed();
  double usElapsed() const { return 4e6 * ((double)cycles() / crystalFrequency()); }
  double realMSec( int c ) { return 4000.0 * ((double)c / crystalFrequency()); }
  
  void setHoldSpeed( bool hold ) { m_hold = hold; }
  
  bool stepRequest() const;
  bool overRequest() const;
  bool finishCallRequest() const;
  bool contRequest() const;
  bool rstRequest() const;
  bool runRequest() const;
  bool quitRequest() const;
  void setStep( bool );
  void setOver( bool );
  void setFinishCall( bool );
  void setCont( bool );
  void setRst( bool );
  void setRun( bool );
  void setQuit( bool );
  
  void setFreeRun( bool fr ) { m_freeRun = fr; }
  void setTracing( bool tr ) { m_tracing = tr; }
  void freeze();
  
  bool isActiveBreakpoint() const;
  
  void wakeUp() { m_command.wakeOne(); }
  
  double callAverage( int pc ) const;
  unsigned lastCall( int pc ) const { return m_callCyclesLast[pc]; }
  unsigned minCall( int pc ) const { return m_callCyclesMin[pc]; }
  unsigned maxCall( int pc ) const { return m_callCyclesMax[pc]; }
  
  // Dirtyflag resides in pic. Everybody has access to pic
  // 
  bool dirty() const { return m_dirty; }
  void setDirty( bool dirty ) { m_dirty = dirty; }
  
  // opcode decoding
  //
  int file( unsigned opcode ) const;
  int literal( unsigned opcode ) const;
  int bit( unsigned opcode ) const;
  int address( unsigned opcode ) const;

  // timers might want to call this
  // 
  void flushInstructionPipeline();
  void fetchInstruction( int );

  bool insideCall() const { return m_callDepth > 0; }
  
  bool animating() const { return m_animating; }
  void setAnimating( bool anim ) { m_animating = anim; }
  
protected:
  QWaitCondition m_command;
  unsigned       m_bitMask[8];
  int            m_w;
  int            m_pc;
  void (Pic::*m_compiledFunction[4096])( unsigned, int );
  void (Pic::*m_instructionPipeline[2])( unsigned, int );
  unsigned       m_opcodePipeline[2];
  int            m_operandPipeline[2];
  int            m_pipelinePos;
  int            m_compiledOperand[4096];
  double         m_callCyclesSum[4096];
  unsigned       m_callCyclesLast[4096];
  unsigned       m_callCyclesMin[4096];
  unsigned       m_callCyclesMax[4096];
  unsigned       m_callCount[4096];
  PicROM        *m_rom;
  PicStack      *m_stack;
  PicRAM        *m_ram;
  SymbolTable   *m_symbolTable;
  PicRegister   *m_sfr[NUM_SFR];
  PicRegister   *m_status;
  PicRegister   *m_pcl;
  PicRegister   *m_pclath;
  PicTimer      *m_timer[NumTimer];
  int            m_numTimer;
  double         m_crystalFrequency;
  double         m_cachedClockSpeed;
  bool           m_hold;
  QTime          m_holdTime;
  unsigned       m_holdCycles;
  double         m_clockFactor;
  mutable QMutex m_mutex;
  DisAss        *m_disassembler;
  InstructionTraceBuffer *m_instructionTraceBuffer;
  RegisterTrace *m_registerTrace;
  int            m_tracePc;
  bool           m_tracing;
  
  bool           m_sleep;
  unsigned       m_cycles;
  //QMutex    m_mutex;
  CodeTable     *m_codeTable;
  int            m_numOpcode;
  
  QTime  m_startTime;         
  unsigned       m_lastCycles;        
  
  bool m_run;
  bool m_step;
  bool m_over;
  bool m_finishCall;
  bool m_cont;
  bool m_rst;
  bool m_quit;
  
  bool m_freeRun;
  bool m_dirty;
  bool m_animating;
  
  int m_overBreakpointPc;
  int m_callDepth;
  
  void push( unsigned );
  unsigned pop();

  void setZ( int );
  void setC( int );
  void setC_Sub( int );
  void setDC( int, int, int );
  void setDC_Sub( int, int, int );
  void setC_Z( int );
  void setC_Z_Sub( int );
  void setC_Z_DC( int, int, int );
  void setC_Z_DC_Sub( int, int, int );
  
  // byte oriented file register operations
  //
  void ADDWF ( unsigned opcode, int operand );
  void ANDWF ( unsigned opcode, int operand );
  void CLRF  ( unsigned opcode, int operand );
  void CLRW  ( unsigned opcode, int operand );
  void COMF  ( unsigned opcode, int operand );
  void DECF  ( unsigned opcode, int operand );
  void DECFSZ( unsigned opcode, int operand );
  void INCF  ( unsigned opcode, int operand );
  void INCFSZ( unsigned opcode, int operand );
  void IORWF ( unsigned opcode, int operand );
  void MOVF  ( unsigned opcode, int operand );
  void MOVWF ( unsigned opcode, int operand );
  void NOP   ( unsigned opcode, int operand );
  void NOP_Internal   ( unsigned opcode, int operand );
  void RLF   ( unsigned opcode, int operand );
  void RRF   ( unsigned opcode, int operand );
  void SUBWF ( unsigned opcode, int operand );
  void SWAPF ( unsigned opcode, int operand );
  void XORWF ( unsigned opcode, int operand );
  
  // bit oriented file register operation
  //
  void BCF   ( unsigned opcode, int operand );
  void BSF   ( unsigned opcode, int operand );
  void BTFSC ( unsigned opcode, int operand );
  void BTFSS ( unsigned opcode, int operand );

  // literal and control operations
  //
  void ADDLW ( unsigned opcode, int operand );
  void ANDLW ( unsigned opcode, int operand );
  void CALL  ( unsigned opcode, int operand );
  void GOTO  ( unsigned opcode, int operand );
  void IORLW ( unsigned opcode, int operand );
  void MOVLW ( unsigned opcode, int operand );
  void RETFIE( unsigned opcode, int operand );
  void RETLW ( unsigned opcode, int operand );
  void RETURN( unsigned opcode, int operand );
  void SLEEP ( unsigned opcode, int operand );
  void SUBLW ( unsigned opcode, int operand );
  void XORLW ( unsigned opcode, int operand );
  void CLRWDT( unsigned opcode, int operand );
  void OPTION( unsigned opcode, int operand );
  void TRIS  ( unsigned opcode, int operand );

  // misc
  //
  void storeResult( int, PicRegister *cell, unsigned opcode );
  void updatePCL();
  
  void createRAM();
  void createTimer();
  void createCodeTable();
  
  void callStatistics( int pc, unsigned cycles );
  
  CodeTable *findCodeFromOpcode( unsigned opcode );
  
  int timeDifference() const;
  
  // The real thing
  //
  bool stepExec();
  void contExec();
  void resetExec();
  
  void incPc();
  void traceInstruction();
  void traceInstruction( int file );
  
  bool executeInstruction();
  bool nextIsNopInternal() const;
  
};

inline bool
Pic::executeInstruction()
{
  // This temporary variable is needed, as the call itself might flush
  // the pipeline. The test would fail then ...
  //
  const bool wasRealInstruction = 
      (m_instructionPipeline[m_pipelinePos] != &Pic::NOP_Internal);
  
  // the function call itself
  //
  (this->*m_instructionPipeline[m_pipelinePos])( m_opcodePipeline[m_pipelinePos], 
                                                 m_operandPipeline[m_pipelinePos] ); 
  
  return wasRealInstruction;
}

inline bool
Pic::nextIsNopInternal() const
{
  return (m_instructionPipeline[(m_pipelinePos ^ 1)] == &Pic::NOP_Internal);
}

inline void 
Pic::flushInstructionPipeline()
{
  // no need to set operand and opcode as it isnt used by NOP
  //
  m_instructionPipeline[0] = m_instructionPipeline[1] = &Pic::NOP_Internal;
}

inline void
Pic::fetchInstruction( int pc )
{
  // fetch instruction
  //
  m_instructionPipeline[m_pipelinePos] = m_compiledFunction[pc];
  m_opcodePipeline[m_pipelinePos]      = m_rom->at( pc );
  m_operandPipeline[m_pipelinePos]     = m_compiledOperand[pc];
  m_pipelinePos ^= 1;
  
  // this takes one cycle
  //
  m_cycles++;
}

inline void
Pic::traceInstruction()
{
  if (!m_tracing) return;
  
  m_instructionTraceBuffer->add( m_tracePc, m_w, m_status->valuePic() );
  m_rom->read( m_tracePc );
}

inline void
Pic::traceInstruction( int file )
{
  if (!m_tracing) return;
  
  m_instructionTraceBuffer->add( m_tracePc, m_w, m_status->valuePic(), file );
  m_rom->read( m_tracePc );
  m_registerTrace->trace();
}

inline void
Pic::callStatistics( int pc, unsigned cycles )
{
  m_callCyclesSum[pc] += cycles;
  m_callCyclesLast[pc] = cycles;
  if (cycles < m_callCyclesMin[pc]) m_callCyclesMin[pc] = cycles;
  if (cycles > m_callCyclesMax[pc]) m_callCyclesMax[pc] = cycles;
  m_callCount[pc]++;
}

inline double 
Pic::callAverage( int pc ) const 
{ 
  return m_callCount[pc] ? 
            (m_callCyclesSum[pc] / (double)m_callCount[pc]) : 
            0.0;
}

inline void
Pic::setPc( int pc, bool setPcl )
{
  m_pc = pc; 
  flushInstructionPipeline();
  fetchInstruction( m_pc );
  
  if (setPcl)
  {
    updatePCL(); 
    m_sfr[PCL]->freeze(); 
  }
}

inline void 
Pic::updatePCL()
{
  // PCH is internal and PCLATH will never be updated
  //
  m_pcl->setValuePic( m_pc ); // will clamp
}

inline int
Pic::pcFromPCL() const
{
  return (m_pcl->valuePic() | ((m_pclath->valuePic() << 8) & 0x7f00));
}

inline void
Pic::incPc()
{
  m_pc++;
  m_pc &= 0x7ff;
  
  updatePCL();
}

inline void 
Pic::setCrystalFrequency( double f ) 
{ 
  m_crystalFrequency = f; 
  m_clockFactor = 1.0 / m_crystalFrequency * 100.0;
}

inline double
Pic::percentSpeed()
{
  return (clockSpeed() * m_clockFactor);
}

inline void
Pic::storeResult( int value, PicRegister *cell, unsigned opcode )
{
  if (opcode & 0x0080)
  {
    // This is a user write
    //
    cell->setValue( value );
  }
  else
  {
    m_w = value & 0xff;
  }
}

// extract from opcode
//
inline int
Pic::address( unsigned opcode ) const
{
  return (opcode & 0x07ff);
}

inline int
Pic::file( unsigned opcode ) const
{
  return (opcode & 0x007f);
}

inline int
Pic::literal( unsigned opcode ) const
{
  return (opcode & 0x00ff);
}

inline int
Pic::bit( unsigned opcode ) const
{
  return ((opcode >> 7) & 0x07);
}

// flag stuff (Try to avoid jumps or conditional stuff)
//
// Below the explanation of the rather obscure bit fiddling.
// If they ere not faster, then they are at least more obscure :o)
//
// Z_BIT: 
//  value (which is 32 bit) is decremented. If it was zero
//  it gets negative (most significant bit is set. Now we
//  shift it at the Z_BIT position and mask (treating as unsigned)
//
//  (Z_BIT fiddling seems to be slower than ((value & 0xff) ? 0 : Z_BIT)
//
// C_BIT:
//  carry occured if bit 8 (0x100) is set. Just shift to it's position
//
// DC_BIT:
//  newValue XOR oldVALUE XOR operand will have bit 0x10 set if DC
//  carry occured.
//  
//  oldValue = 00001111
//  operand  = 00000001
//  opration = ADD
//  newValue = 00010000
//
//  oldValue XOR newValue = 00011111
//    XOR operand = 00011110 -> & 00010000 != 0 OK
//
//  oldValue =  10001000
//  operand  =  10001000
//  opration = ADD
//  newValue = 100010000
//
//  oldValue XOR newValue = 110011000
//    XOR operand = 100010000 -> & 00010000 != 0 OK
//
//
//  oldValue =  10000000
//  operand  =  10000000
//  opration = ADD
//  newValue = 100000000
//
//  oldValue XOR newValue = 110000000
//    XOR operand = 100000000 -> & 00010000 == 0 OK
//
// carry's are inverted for substraction by xoring with itself
//
inline void
Pic::setZ( int value )
{
  m_status->setValuePic( (m_status->valuePic() & ~Z_BIT) |
//                         ((((unsigned)((value & 0x000000ff)-1)) >> 29) & Z_BIT) );
                         ((value & 0xff) ? 0 : Z_BIT) );
}

inline void
Pic::setC( int value )
{
  m_status->setValuePic( (m_status->valuePic() & ~C_BIT) |
                         ((value & 0x00000100) >> 8) );
//                         ((value & 0x100) ? C_BIT : 0) );
}

inline void
Pic::setC_Z( int value )
{
  m_status->setValuePic( (m_status->valuePic() & ~(C_BIT | Z_BIT)) |
                         ((value & 0x00000100) >> 8) |
//                         ((value & 0x100) ? C_BIT : 0) |
//                         (((unsigned)((value & 0x000000ff)-1) >> 29) & Z_BIT) );
                         ((value & 0xff) ? 0 : Z_BIT) );
}

inline void
Pic::setC_Z_Sub( int value )
{
  m_status->setValuePic( (m_status->valuePic() & ~(C_BIT | Z_BIT)) |
                         (((value & 0x00000100) >> 8) ^ C_BIT) |
//                         ((value & 0x100) ? 0 : C_BIT) |
//                         (((unsigned)((value & 0x000000ff)-1) >> 29) & Z_BIT) );
                         ((value & 0xff) ? 0 : Z_BIT) );
}

inline void
Pic::setC_Z_DC( int oldValue, int value, int operand )
{
  m_status->setValuePic( (m_status->valuePic() & ~(C_BIT | Z_BIT | DC_BIT)) |
                         ((value & 0x00000100) >> 8) |
//                         ((value & 0x100) ? C_BIT : 0) |
//                         (((unsigned)((value & 0x000000ff)-1) >> 29) & Z_BIT)|
                         ((value & 0xff) ? 0 : Z_BIT) |
                         (((oldValue ^ value ^ operand) & 0x00000010) >> 3) );
//                         (((oldValue ^ value ^ operand) & 0x010) ? DC_BIT : 0) );
//                         ((value & 0x010) && !(oldValue & 0x010) ? DC_BIT : 0) );
}

inline void
Pic::setC_Z_DC_Sub( int oldValue, int value, int operand )
{
  m_status->setValuePic( (m_status->valuePic() & ~(C_BIT | Z_BIT | DC_BIT)) |
                         (((value & 0x00000100) >> 8) ^ C_BIT) |
//                         ((value & 0x100) ? 0 : C_BIT) |
//                         (((unsigned)((value & 0x000000ff)-1) >> 29) & Z_BIT) |
                         ((value & 0xff) ? 0 : Z_BIT) |
                         ((((oldValue ^ value ^ operand) & 0x00000010) >> 3) ^ DC_BIT) );
//                         (((oldValue ^ value ^ operand) & 0x010) ? 0 : DC_BIT) );
//                         ((value & 0x010) && !(oldValue & 0x010) ? 0 : DC_BIT) );
}

inline void
Pic::setC_Sub( int value )
{
  m_status->setValuePic( (m_status->valuePic() & ~C_BIT) |
                         (((value & 0x00000100) >> 8) ^ C_BIT) );
//                         ((value & 0x100) ? 0 : C_BIT) );
}

inline void
Pic::setDC( int oldValue, int value, int operand )
{
  m_status->setValuePic( (m_status->valuePic() & ~DC_BIT) |
                         (((oldValue ^ value ^ operand) & 0x00000010) >> 3 ));
//                         (((oldValue ^ value ^ operand) & 0x010) ? DC_BIT : 0) );
//                         ((value & 0x010) && !(oldValue & 0x010) ? DC_BIT : 0) );
}

inline void
Pic::setDC_Sub( int oldValue, int value, int operand )
{
  m_status->setValuePic( (m_status->valuePic() & ~DC_BIT) |
                         ((((oldValue ^ value ^ operand) & 0x00000010) >> 3) ^ DC_BIT) );
//                         (((oldValue ^ value ^ operand) & 0x010) ? 0 : DC_BIT) );
//                         ((value & 0x010) && !(oldValue & 0x010) ? 0 : DC_BIT) );
}

inline void
Pic::setStatusFlag( int flag, bool on )
{
  if (on)
  {
    m_status->setBitPic( flag );
  }
  else
  {
    m_status->clearBitPic( flag );
  }
}

#endif // PIC_HH
