//======================================================================
// File:		picregister.h
// Author:	Matthias Toussaint
// Created:	Mon Jul 26 23:08:00 CET 2004
//----------------------------------------------------------------------
// Permission to use, copy, modify, and distribute this software and its
// documentation  for any  purpose and  without fee is  hereby  granted,
// provided  that below copyright notice appear  in all copies  and that
// both  that  copyright  notice and  this permission  notice  appear in
// supporting documentation.
// 
// This  file is  provided AS IS  with no  warranties  of any kind.  The
// author shall  have no liability  with respect  to the infringement of
// copyrights, trade  secrets  or any patents by  this file  or any part
// thereof.  In no event will the author be liable  for any lost revenue
// or profits or other special, indirect and consequential damages.
//----------------------------------------------------------------------
// (c) 2004 Matthias Toussaint
//======================================================================

#ifndef PICREGISTER_HH
#define PICREGISTER_HH

#include <qstring.h>

#include <watchcondition.h>

#include <iostream>

/** This is the representation of _one_byte_ in the RAM
*/
class PicRegister
{
public:
  enum Type
  {
    GeneralPurpose = 0,
    Null,
    SFR,
    STATUS,
    FSR,
    INDF
  };

  PicRegister( unsigned addr );
  PicRegister( Type type=GeneralPurpose, const QString & name=QString::null,
               int mask=0xff, int resetValue=0x00 );
  virtual ~PicRegister();
  
  bool isSFR() const { return (m_type > Null); }
  bool isGeneralPurpose() const { return m_type == GeneralPurpose; }
  bool isNull() const { return m_type == Null; }
  
  virtual void reset();
  
  QString name() const { return m_name; }
  void setName( const QString & name ) { m_name = name; }
  void createName( int );
  
  virtual int value() const;
  virtual int valuePic() const;
  virtual void setBit( int bit );
  virtual void clearBit( int bit );
  virtual bool bitSet( int bit ) const;
  
  virtual void setValue( int value );
  virtual void setValuePic( int value );
  virtual void setBitPic( int bit );
  virtual void clearBitPic( int bit );
  virtual bool bitSetPic( int bit ) const;
  
  QString flagName( int i ) const { return m_flagName[i]; }
  void setFlagNames( const QString & f0, const QString & f1,
                     const QString & f2, const QString & f3,
                     const QString & f4, const QString & f5,
                     const QString & f6, const QString & f7 );
  
  Type type() const { return m_type; }
  int mask() const { return m_mask; }
  bool hasBit( int i ) { return (m_mask & m_bitMask[i]); }
  
  int freezedValue() const { return m_freezedValue; }
  int freezedReadCnt() const { return m_freezedReadCnt; }
  int freezedWriteCnt() const { return m_freezedWriteCnt; }
  bool freezedBitSet( int bit ) { return m_freezedValue & m_bitMask[bit]; }
  virtual void freeze();
  
  bool traced() const { return m_traced; }
  void setTraced( bool traced ) { m_traced = traced; }
  bool watched() const { return m_watchCondition; }
  void setWatchCondition( WatchCondition *watchCondition ) 
      { m_watchCondition = watchCondition; }
  
  // TODO: put this flag into ram
  static bool watchActive;

protected:
  QString               m_name;
  QString               m_flagName[8];
  int                   m_value;
  int                   m_freezedValue;
  const Type            m_type;
  int                   m_mask;
  int                   m_resetValue;
  WatchCondition       *m_watchCondition;
  mutable unsigned long m_readCnt;
  unsigned long         m_writeCnt;
  unsigned long         m_freezedReadCnt;
  unsigned long         m_freezedWriteCnt;
  bool                  m_traced;
  
  static int m_bitMask[8];
    
  void checkWatch( WatchCondition::AccessMode accessMode ) const;
  
private:
  void createDefaultFlagNames();

};

inline void
PicRegister::reset()
{
  setValuePic( m_resetValue );
  m_readCnt = m_writeCnt = 0;
}

inline void 
PicRegister::freeze()
{
  m_freezedValue    = m_value; 
  m_freezedReadCnt  = m_readCnt;//  - m_freezedReadCnt;
  m_freezedWriteCnt = m_writeCnt;// - m_freezedWriteCnt;
}

inline bool
PicRegister::bitSetPic( int bit ) const
{
  return m_value & m_bitMask[bit];
}

inline bool
PicRegister::bitSet( int bit ) const
{
  checkWatch( WatchCondition::Read );
  
  return m_value & m_bitMask[bit];
}
 
inline void
PicRegister::setBit( int bit )
{
  m_value |= m_bitMask[bit] & m_mask;
  m_writeCnt++;
  
  checkWatch( WatchCondition::Write );
}

inline void
PicRegister::clearBit( int bit )
{
  m_value &= ~m_bitMask[bit];
  m_writeCnt++;
  
  checkWatch( WatchCondition::Write );
}

inline void
PicRegister::setBitPic( int bit )
{
  m_value |= m_bitMask[bit] & m_mask;
  
  checkWatch( WatchCondition::InternalWrite );
}

inline void
PicRegister::clearBitPic( int bit )
{
  m_value &= ~m_bitMask[bit];
  
  checkWatch( WatchCondition::InternalWrite );
}

inline int
PicRegister::valuePic() const
{
  return m_value;
}

inline void
PicRegister::setValuePic( int value )
{
  m_value = value & m_mask; 
  
  checkWatch( WatchCondition::InternalWrite );
}

inline void
PicRegister::setValue( int value )
{
  m_value = value & m_mask; 
  
  checkWatch( WatchCondition::Write );
  
  m_writeCnt++;
}

inline int
PicRegister::value() const
{
  checkWatch( WatchCondition::Read );
  
  m_readCnt++;  
  
  return m_value;
}
 
inline void
PicRegister::checkWatch( WatchCondition::AccessMode accessMode ) const
{
  if (m_watchCondition && 
      m_watchCondition->match( m_value, accessMode ))
  {
    PicRegister::watchActive = true;
  }
}

#endif // PICREGISTER_HH
