//======================================================================
// File:		ramview.cpp
// Author:	Matthias Toussaint
// Created:	Mon Jul 26 23:08:00 CET 2004
//----------------------------------------------------------------------
// Permission to use, copy, modify, and distribute this software and its
// documentation  for any  purpose and  without fee is  hereby  granted,
// provided  that below copyright notice appear  in all copies  and that
// both  that  copyright  notice and  this permission  notice  appear in
// supporting documentation.
// 
// This  file is  provided AS IS  with no  warranties  of any kind.  The
// author shall  have no liability  with respect  to the infringement of
// copyrights, trade  secrets  or any patents by  this file  or any part
// thereof.  In no event will the author be liable  for any lost revenue
// or profits or other special, indirect and consequential damages.
//----------------------------------------------------------------------
// (c) 2004 Matthias Toussaint
//======================================================================

#include <qcursor.h>
#include <qheader.h>
#include <qlabel.h>
#include <qlineedit.h>
#include <qpopupmenu.h>
#include <highlightcondition.h>
#include <qtable.h>

#include <picregistertableitem.h>
#include <ramview.h>
#include <stringconversion.h>

#include <stdlib.h>

#include <iostream>

RAMView::RAMView( QWidget *parent, const char *name ) :
  UIRAMView( parent, name ),
  m_pic( 0 ),
  m_currentReg( 0 )
{
  m_edit[0] = ui_red;
  m_edit[1] = ui_green;
  m_edit[2] = ui_blue;
  m_edit[3] = ui_cyan;
  m_edit[4] = ui_magenta;
  m_edit[5] = ui_yellow;

  m_colors[0] = Qt::red.light();
  m_colors[1] = Qt::green;
  m_colors[2] = Qt::blue.light();
  m_colors[3] = Qt::cyan;
  m_colors[4] = Qt::magenta.light();
  m_colors[5] = Qt::yellow;
  
  int hw = fontMetrics().width( "DD" ) + 8;
  
  QHeader *hheader = ui_table->horizontalHeader();
  hheader->setResizeEnabled( false );
  hheader->setClickEnabled( false );
  for (int i=0; i<16; ++i)
  {
    hheader->resizeSection( i, hw );
  }
  
  QHeader *vheader = ui_table->verticalHeader();
  vheader->setResizeEnabled( false );
  vheader->setClickEnabled( false );
  
  connect( ui_table, SIGNAL( currentChanged( int, int ) ),
           this, SLOT( currentChangedSLOT( int, int ) ) );
  
  connect( ui_hex, SIGNAL( textChanged( const QString & ) ),
           this, SLOT( hexChangedSLOT( const QString & ) ) );
  connect( ui_dec, SIGNAL( textChanged( const QString & ) ),
           this, SLOT( decChangedSLOT( const QString & ) ) );
  connect( ui_bin, SIGNAL( textChanged( const QString & ) ),
           this, SLOT( binChangedSLOT( const QString & ) ) );
  
  connect( ui_red, SIGNAL( textChanged( const QString & ) ),
           this, SLOT( redChangedSLOT( const QString & ) ));
  connect( ui_green, SIGNAL( textChanged( const QString & ) ),
           this, SLOT( greenChangedSLOT( const QString & ) ));
  connect( ui_blue, SIGNAL( textChanged( const QString & ) ),
           this, SLOT( blueChangedSLOT( const QString & ) ));
  connect( ui_cyan, SIGNAL( textChanged( const QString & ) ),
           this, SLOT( cyanChangedSLOT( const QString & ) ));
  connect( ui_magenta, SIGNAL( textChanged( const QString & ) ),
           this, SLOT( magentaChangedSLOT( const QString & ) ));
  connect( ui_yellow, SIGNAL( textChanged( const QString & ) ),
           this, SLOT( yellowChangedSLOT( const QString & ) ));
  
  ui_redLabel->setBackgroundColor( m_colors[0] );
  ui_greenLabel->setBackgroundColor( m_colors[1] );
  ui_blueLabel->setBackgroundColor( m_colors[2] );
  ui_cyanLabel->setBackgroundColor( m_colors[3] );
  ui_magentaLabel->setBackgroundColor( m_colors[4] );
  ui_yellowLabel->setBackgroundColor( m_colors[5] );
  
  m_lastValue = new int [4*128];
  for (int i=0; i<4*128; ++i) m_lastValue[i] = -1;
  
  m_popup = new QPopupMenu( this );
  connect( m_popup, SIGNAL( activated( int ) ),
           this, SLOT( popupSLOT( int ) ));
  
  connect( ui_table, SIGNAL( pressed( int, int, int, const QPoint & ) ),
           this, SLOT( pressedSLOT( int, int, int, const QPoint & ) ));
  
  //ui_table->viewport()->setBackgroundMode( NoBackground );
}

RAMView::~RAMView()
{
  delete [] m_lastValue;
  delete [] m_condition;
}

void
RAMView::resizeEvent( QResizeEvent *ev )
{
  UIRAMView::resizeEvent( ev );
  
  const double hw = (double)ui_table->visibleWidth() / 16.0;
  
  ui_table->setUpdatesEnabled( false );
  
  ui_table->setColumnWidth( 0, (int)hw );
  
  for (int i=1; i<16; ++i)
  {
    const int x1 = (int)((double)(i-1)*hw);
    const int x2 = (int)((double)i*hw);
    
    ui_table->setColumnWidth( i, x2-x1 );
  }
  
  ui_table->setUpdatesEnabled( true );
  ui_table->viewport()->update();
}

void
RAMView::show()
{
  UIRAMView::show();
  refreshSLOT();
}

void
RAMView::setEditable( bool editable )
{
  ui_table->setReadOnly( !editable );
  ui_hex->setReadOnly( !editable );
  ui_dec->setReadOnly( !editable );
  ui_bin->setReadOnly( !editable );
  ui_ascii->setReadOnly( !editable );
}

void
RAMView::registerChangedSLOT( PicRegister * )
{
  refreshSLOT();
}

void
RAMView::setPic( Pic * pic )
{
  m_pic = pic;

  if (!m_pic)
  {
    ui_table->setNumRows(0);
    return;
  }

  m_condition = new HighlightCondition [6];
  
  for (int i=0; i<6; ++i) m_condition[i].setRAM( pic->ram() );
  
  QString str;
  
  ui_table->setNumRows( pic->ram()->numBanks()*8 );
  
  QHeader *vheader = ui_table->verticalHeader();
  
  for (int i=0; i<ui_table->numRows(); ++i)
  {
    str.sprintf( "%03X", 16*i );
    vheader->setLabel( i, str );
  }
  

  for (int i=0; i<pic->ram()->numBanks()*128; ++i)
  {
    int bank = (i>>7) & 0x03;
    int cell = i & 0x7f;
    
    PicRegister *reg = m_pic->ram()->get( bank, cell );
    
    if (reg)
    {
      ui_table->setItem( i/16, i%16, 
                         new PicRegisterTableItem( ui_table, 
                                                   reg, 
                                                   m_pic->ram()->isShadow( bank, cell ) ));
    }
    else
    {
      ui_table->setItem( i/16, i%16, 
                         new PicRegisterTableItem( ui_table, 0, true ) );
    }
  }
}

void
RAMView::refreshSLOT()
{
  static QString str;
  
  //std::cerr << "RAMView::refreshSLOT()" << std::endl;
  if (!m_pic) return;
  if (!isVisible()) return;
  
  //ui_table->viewport()->setUpdatesEnabled( false );
  
  const int numBanks = m_pic->ram()->numBanks();
  
  for (int bank=0; bank<numBanks; bank++)
  {
    for (int id=0; id<128; id++)
    {
      PicRegister *reg = m_pic->ram()->rawGet( bank, id );

      const int i = id + (bank << 7);
      
      if (reg && reg->freezedValue() != m_lastValue[i])
      {
        ui_table->setText( i >> 4, i & 0x0f, 
            StringConversion::hexString( reg->freezedValue() ) );

        m_lastValue[i] = reg->freezedValue();

        if (reg == m_currentReg)
        {
          currentChangedSLOT( m_currentReg );
        }
      }
    }
  }
  
  setColors();
  
  //ui_table->viewport()->setUpdatesEnabled( true );
  //ui_table->viewport()->update();
}

void
RAMView::currentChangedSLOT( PicRegister *reg )
{
  static QString str;
  
  //std::cerr << "RAMView::currentChangedSLOT( PicRegister *reg )" << std::endl;
  if (!reg) return;
 
  ui_name->setText( reg->name() );
  
  ui_hex->blockSignals( true );
  ui_dec->blockSignals( true );
  ui_bin->blockSignals( true );
  ui_ascii->blockSignals( true );  
  
  if (!ui_hex->hasFocus() )
  {
    ui_hex->setText( StringConversion::hexString( reg->freezedValue() ) );
  }
  
  if (!ui_dec->hasFocus() )
  {
    ui_dec->setText( StringConversion::decString( reg->freezedValue() ) );
  }
  
  if (!ui_bin->hasFocus() )
  {
    ui_bin->setText( StringConversion::binString( reg->freezedValue() ) );
  }
  
  if (!ui_ascii->hasFocus() )
  {
    QChar c( reg->freezedValue() );
    
    str.sprintf( "%c", c.isLetterOrNumber() ? (char)c.latin1() : ' ' );
    
    ui_ascii->setText( str );
  }
  
  ui_hex->blockSignals( false );
  ui_dec->blockSignals( false );
  ui_bin->blockSignals( false );
  ui_ascii->blockSignals( false );
}

void
RAMView::currentChangedSLOT( int row, int col )
{
  //std::cerr << "RAMView::currentChangedSLOT( int row, int col )" << std::endl;
  const int addr = (row << 4) + col;
  const int bank = addr >> 7;
  const int id   = addr & 0x7f;
  
  m_currentReg = m_pic->ram()->get( bank, id );
  currentChangedSLOT( m_currentReg );
}

void
RAMView::registerSelectedSLOT( int addr )
{
  const int row = addr >> 4;
  const int col = addr & 0x0F;
  
  ui_table->setCurrentCell( row, col );
}

void
RAMView::setValue( int value )
{
  //std::cerr << "RAMView::setValue( int value )" << std::endl;
  m_currentReg->setValuePic( value );
  m_currentReg->freeze();
  
  emit updateViews();
  
  //currentChangedSLOT( m_currentReg );
}

void
RAMView::hexChangedSLOT( const QString & text )
{
  valueChanged( text, 16 );
}

void
RAMView::decChangedSLOT( const QString & text )
{
  valueChanged( text, 10 );
}

void
RAMView::binChangedSLOT( const QString & text )
{
  valueChanged( text, 2 );
}

void
RAMView::valueChanged( const QString & text, int base )
{
  if (m_currentReg)
  {
    setValue( text.toInt( 0, base ) );
    setColors();
  }
}

void 
RAMView::redChangedSLOT( const QString & pattern )
{
  colorChanged( pattern, ColorRed );
}

void 
RAMView::greenChangedSLOT( const QString & pattern )
{
  colorChanged( pattern, ColorGreen );
}

void 
RAMView::blueChangedSLOT( const QString & pattern )
{
  colorChanged( pattern, ColorBlue );
}

void 
RAMView::cyanChangedSLOT( const QString & pattern )
{
  colorChanged( pattern, ColorCyan );
}

void 
RAMView::magentaChangedSLOT( const QString & pattern )
{
  colorChanged( pattern, ColorMagenta );
}

void 
RAMView::yellowChangedSLOT( const QString & pattern )
{
  colorChanged( pattern, ColorYellow );
}

void
RAMView::colorChanged( const QString & pattern, ColorItem item )
{
  if (!m_pic) return;
  
  m_condition[item].setPattern( pattern );
  setColors();
  
  if (m_condition[item].syntaxOk())
  {
    m_edit[item]->setBackgroundColor( colorGroup().base() );
  }
  else
  {
    int red   = colorGroup().base().red();
    int green = colorGroup().base().green() - 20;
    int blue  = colorGroup().base().blue() - 20;
    
    if (green < 0) green = 0;
    if (blue < 0)  blue = 0;
    
    m_edit[item]->setBackgroundColor( QColor( red, green, blue ) );
  }
}

void
RAMView::setColors()
{
  if (!m_pic) return;
  
  for (int col=0; col<ui_table->numCols(); ++col)
  {
    for (int row=0; row<ui_table->numRows(); ++row)
    {
      PicRegisterTableItem *item = 
          (PicRegisterTableItem *)ui_table->item( row, col );
      
      if (!item->isNullRegister())
      {
        bool updateCell = false;

        if (item->isSpecialColor())
        {
          item->resetSpecialColor();
          updateCell = true;
        }

        const int addr = row*16+col;

        for (int color=0; color<6; ++color)
        {
          if (m_condition[color].match( addr, item->value()))
          {
            item->setSpecialColor(m_colors[color]);
            updateCell = true;

            break;
          }
        }

        if (updateCell) ui_table->updateCell( row, col );
      }
    }
  }
}

void
RAMView::pressedSLOT( int row, int col, int button, const QPoint & )
{
  if (RightButton == button)
  {
    ui_table->setCurrentCell( row, col );
    currentChangedSLOT( row, col );
    
    if (!m_currentReg->isNull())
    {        
      m_popup->clear();
      QString text;
      
      if (m_currentReg->watched())
      {
        text.sprintf( tr("Remove %s from Watch"), m_currentReg->name().latin1() );
        m_popup->insertItem( text, REMOVE_WATCH_ID );
      }
      else
      {
        text.sprintf( tr("Add %s to Watch"), m_currentReg->name().latin1() );
        m_popup->insertItem( text, ADD_WATCH_ID );
      }
      
      if (m_currentReg->traced())
      {
        text.sprintf( tr("Remove %s from Trace"), m_currentReg->name().latin1() );
        m_popup->insertItem( text, REMOVE_TRACE_ID );
      }
      else
      {
        text.sprintf( tr("Add %s to Trace"), m_currentReg->name().latin1() );
        m_popup->insertItem( text, ADD_TRACE_ID );
      }
      
      m_popup->popup( QCursor::pos() );
    }
  } 
}

void
RAMView::popupSLOT( int id )
{
  switch (id)
  {
  case REMOVE_WATCH_ID:
    emit removeWatch( m_currentReg );
    break;
  case ADD_WATCH_ID:
    emit addWatch( m_currentReg );
    break;
  case REMOVE_TRACE_ID:
    emit removeTrace( m_currentReg );
    break;
  case ADD_TRACE_ID:
    emit addTrace( m_currentReg );
    break;
  }
}
