//======================================================================
// File:		romviewtable.cpp
// Author:	Matthias Toussaint
// Created:	Mon Jul 26 23:08:00 CET 2004
//----------------------------------------------------------------------
// Permission to use, copy, modify, and distribute this software and its
// documentation  for any  purpose and  without fee is  hereby  granted,
// provided  that below copyright notice appear  in all copies  and that
// both  that  copyright  notice and  this permission  notice  appear in
// supporting documentation.
// 
// This  file is  provided AS IS  with no  warranties  of any kind.  The
// author shall  have no liability  with respect  to the infringement of
// copyrights, trade  secrets  or any patents by  this file  or any part
// thereof.  In no event will the author be liable  for any lost revenue
// or profits or other special, indirect and consequential damages.
//----------------------------------------------------------------------
// (c) 2004 Matthias Toussaint
//======================================================================

#include <qcursor.h>
#include <qdrawutil.h>
#include <qpixmap.h>
#include <qpopupmenu.h>

#include <math.h>

#include <pic.h>
#include <romviewtable.h>

#include <bp.xpm>
#include <ibp.xpm>
#include <cbp.xpm>
#include <icbp.xpm>
#include <pc.xpm>

ROMViewTableHeader::ROMViewTableHeader( QWidget *parent, const char *name ) :
  QHeader( parent, name )
{
  addLabel( tr("PC") );
  addLabel( tr("Opcode") );
  addLabel( tr("Mnemonic") );
  addLabel( tr("Executed") );
  addLabel( tr("Av. Cycles") );
  addLabel( tr("Last") );
  addLabel( tr("Min") );
  addLabel( tr("Max") );
  
  m_popup = new QPopupMenu( this );
  m_popup->setCheckable( true );
  for (int i=0; i<NumSections; ++i)
  {
    m_popup->insertItem( label(i), i );
    m_popup->setItemChecked( i, true );
  }
  connect( m_popup, SIGNAL( activated(int) ),
           this, SLOT( popupSLOT( int ) ));
  
}

ROMViewTableHeader::~ROMViewTableHeader()
{
}

void
ROMViewTableHeader::popupSLOT( int id )
{
  bool toggle = !m_popup->isItemChecked( id );
  m_popup->setItemChecked( id, toggle );
  
  emit setVisibility( id, toggle );
}

void
ROMViewTableHeader::mousePressEvent( QMouseEvent *ev )
{
  if (ev->button() == RightButton)
  {
    m_popup->popup( QCursor::pos() );
  }
  else
  {
    QHeader::mousePressEvent( ev );
  }
}

ROMViewTable::ROMViewTable( QWidget *parent, const char *name ) :
  QScrollView( parent, name ),
  m_numLines( 0 ),
  m_current( 0 )
{
  viewport()->setBackgroundMode( QWidget::NoBackground );
  
  m_bpPix   = new QPixmap((const char **)bp_xpm);
  m_ibpPix  = new QPixmap((const char **)ibp_xpm);
  m_cbpPix  = new QPixmap((const char **)cbp_xpm);
  m_icbpPix = new QPixmap((const char **)icbp_xpm);
  m_pcPix   = new QPixmap((const char **)pc_xpm);
  
  m_popup = new QPopupMenu( this );
  
  connect( m_popup, SIGNAL( activated( int ) ),
           this, SLOT( popupSLOT( int ) ));
  
  m_header = new ROMViewTableHeader( this );
  m_header->move( 2, 2 );
  m_header->setTracking( true );
  
  connect( m_header, SIGNAL( sizeChange(int, int, int) ),
           this, SLOT( headerResizedSLOT() ));
  connect( m_header, SIGNAL( indexChange(int, int, int) ),
           this, SLOT( headerResizedSLOT() ));
  
  setMargins( 0, m_header->height(), 0, 0 );
  
  setStatisticsColorsSLOT( Qt::green.light( 180 ),
                           Qt::blue.light( 180 ),
                           Qt::yellow.light( 180 ),
                           Qt::red.light( 180 ) );
  
  setHScrollBarMode( QScrollView::AlwaysOff );
  
  setFixedFontSLOT( QFont( "Fixed", 12 ) );  
  
  for (int i=0; i<ROMViewTableHeader::NumSections; ++i)
  {
    setColumnVisibilitySLOT( i, true );
  }
  
  connect( m_header, SIGNAL( setVisibility( int, bool ) ),
           this, SLOT( setColumnVisibilitySLOT( int, bool ) ));
}

ROMViewTable::~ROMViewTable()
{
  delete m_bpPix;   
  delete m_ibpPix;  
  delete m_cbpPix;  
  delete m_icbpPix; 
  delete m_pcPix;
}

void
ROMViewTable::setStatisticsColorsSLOT( const QColor & low, 
                                       const QColor & medium,
                                       const QColor & high, 
                                       const QColor & highest )
{
  m_brush[3] = low;
  m_brush[2] = medium;
  m_brush[1] = high;
  m_brush[0] = highest;
  
  m_brushDark[3] = low.dark( 105 );
  m_brushDark[2] = medium.dark( 105 );
  m_brushDark[1] = high.dark( 105 );
  m_brushDark[0] = highest.dark( 105 );
  
  viewport()->update();
}

void
ROMViewTable::setColumnVisibilitySLOT( int col, bool show )
{
  m_visible[col] = show;
  
  QFontMetrics fm( m_fixedFont );
  
  switch (col)
  {
  case ROMViewTableHeader::PC:
    m_header->resizeSection( col, 
                             m_visible[col] ? 
                             fontMetrics().width( "000" ) + 48 : 0 );
    break;
  case ROMViewTableHeader::OPCODE:
    m_header->resizeSection( col, 
                             m_visible[col] ? 
                             fm.width( "0000 : 00 0000 0000 0000" ) + 8 : 0 );
    break;
  case ROMViewTableHeader::MNEMONIC:
    m_header->resizeSection( col, 
                             m_visible[col] ? 
                             fm.width( "0000000000000000000" ) + 8 : 0 );
    break;
  case ROMViewTableHeader::READ:
    m_header->resizeSection( col, 
                             m_visible[col] ? 
                             fm.width( "0000000000" ) + 8 : 0 );
    break;
  case ROMViewTableHeader::AVG:
    m_header->resizeSection( col, 
                             m_visible[col] ? 
                             fm.width( "00000000" ) + 8 : 0 );
    break;
  case ROMViewTableHeader::LAST:
    m_header->resizeSection( col, 
                             m_visible[col] ? 
                             fm.width( "0000" ) + 8 : 0 );
    break;
  case ROMViewTableHeader::MIN:
    m_header->resizeSection( col, 
                             m_visible[col] ? 
                             fm.width( "0000" ) + 8 : 0 );
    break;
  case ROMViewTableHeader::MAX:
    m_header->resizeSection( col, 
                             m_visible[col] ? 
                             fm.width( "0000" ) + 8 : 0 );
    break;
  }
  
  headerResizedSLOT();
}

void
ROMViewTable::setFixedFontSLOT( const QFont & font )
{
  m_fixedFont = font;
  QFontMetrics fm( m_fixedFont );
  m_header->resizeSection( ROMViewTableHeader::PC, fm.width( "000" ) + 48 );
  m_header->resizeSection( ROMViewTableHeader::OPCODE, fm.width( "0000 : 00 0000 0000 0000" ) + 8 );
  m_header->resizeSection( ROMViewTableHeader::MNEMONIC, fm.width( "0000000000000000000" ) + 8 );
  
  headerResizedSLOT();
}

void
ROMViewTable::setPic( Pic * pic )
{
  m_pic = pic;
  m_numLines = m_pic->rom()->size();
  resizeContents( visibleWidth(), m_numLines*(fontMetrics().height()+4) );
}

void
ROMViewTable::setLine( int line, const QString & opcode, const QString & mnemonic )
{
  m_opcode[line] = opcode;
  m_mnemonic[line] = mnemonic;
}

int
ROMViewTable::lineHeight() const
{
  QFontMetrics ffm( m_fixedFont );
  return QMAX( fontMetrics().height(), ffm.height() ) + 2;
}

void
ROMViewTable::resizeEvent( QResizeEvent *ev )
{
  QScrollView::resizeEvent( ev );
  QFontMetrics nfm = fontMetrics();
  m_header->resize( QMIN( visibleWidth(), m_header->headerWidth() ), 
                    nfm.height()+8 );
  setMargins( 0, m_header->height(), 0, 0 );
}

void
ROMViewTable::headerResizedSLOT()
{
  m_header->resize( QMIN( visibleWidth(), m_header->headerWidth() ), 
                    fontMetrics().height()+8 );
  setMargins( 0, m_header->height(), 0, 0 );
  
  resizeContents( m_header->headerWidth(), 
                  m_numLines*lineHeight() );
  viewport()->update();
}

void
ROMViewTable::drawContents( QPainter *p, int cx, int cy, int cw, int ch )
{
  const QColorGroup cg = colorGroup();
  
  const QRect clipRect( cx, cy, cw, ch );
  
  // erase unused area
  //
  if (visibleWidth() > m_header->headerWidth())
  {
    p->fillRect( m_header->headerWidth(), cy,
                 visibleWidth()-m_header->headerWidth(), ch,
                 cg.background() );
  }

  const int lh = lineHeight();
  const int startLine = cy/lh;
  const int lastLine = (cy+ch) / lh + 1;
  
  const QBrush highlight     = Qt::darkGreen;
  const QColor highlightText = Qt::white;
    
  const QBrush bgBase     = cg.base();
  const QBrush bgBaseDark = cg.base().dark(105);
    
  const QBrush bgNormal = cg.background();
  const QBrush bgDark   = cg.background().dark(105);
  
  
  //m_pic->lock();
  
  const int pc = m_pic->pc();
  const unsigned long maxRead = m_pic->rom()->maxRead();
      
  for (int i=startLine; i<m_numLines && i<lastLine; ++i)
  {
    const double        avg      = m_pic->callAverage( i );    
    const unsigned long readCnt  = m_pic->rom()->readCntAt( i );
    const int y = i*lh;      
    
    m_tmpText.sprintf( "%03X", i );
       
    p->save();
    p->setFont( m_fixedFont );
    
    // draw code stuff
    //
    if (pc == i)
    {
      p->setPen( highlightText );
      
      if (m_visible[ROMViewTableHeader::PC])
      {
        drawText( p, clipRect, ROMViewTableHeader::PC, y, lh, 
                  AlignRight | AlignVCenter, m_tmpText, highlight );
        p->drawPixmap( m_header->sectionPos( ROMViewTableHeader::PC )+2, 
                       i*lh+(lh-13)/2, *m_pcPix );
      }
      if (m_visible[ROMViewTableHeader::OPCODE])
      {
        drawText( p, clipRect, ROMViewTableHeader::OPCODE, y, lh, 
                  AlignLeft | AlignVCenter, m_opcode[i], highlight );
      }
      if (m_visible[ROMViewTableHeader::MNEMONIC])
      {
        drawText( p, clipRect, ROMViewTableHeader::MNEMONIC, y, lh, 
                  AlignLeft | AlignVCenter, m_mnemonic[i], highlight );
      }      
    }
    else
    {
      const QBrush bgBrush = ( (i&1) ? bgBaseDark : bgBase );
      
      if (m_visible[ROMViewTableHeader::PC])
      {
        drawText( p, clipRect, ROMViewTableHeader::PC, y, lh, 
                  AlignRight | AlignVCenter, m_tmpText, bgBrush );
      }
      if (m_visible[ROMViewTableHeader::OPCODE])
      {
        drawText( p, clipRect, ROMViewTableHeader::OPCODE, y, lh, 
                  AlignLeft | AlignVCenter, m_opcode[i], bgBrush );
      }
      if (m_visible[ROMViewTableHeader::MNEMONIC])
      {
        drawText( p, clipRect, ROMViewTableHeader::MNEMONIC, y, lh, 
                  AlignLeft | AlignVCenter, m_mnemonic[i], bgBrush );
      }
    }
 
    p->restore();   
    
    // draw breakpoint pixmap (if any)
    //
    if (m_pic->rom()->isBreakpoint(i) && 
        m_visible[ROMViewTableHeader::PC] &&
        !m_pic->rom()->temporaryBreakpoint(i))
    {
      QPixmap *pix;
      
      if (m_pic->rom()->breakpointEnabled(i))
      {
        if (m_pic->rom()->breakpointConditional(i))
        {
          pix = m_cbpPix;
        }
        else
        {
          pix = m_bpPix;
        }
      }
      else
      {
        if (m_pic->rom()->breakpointConditional(i))
        {
          pix = m_icbpPix;
        }
        else
        {
          pix = m_ibpPix;
        }
      }  
      
      p->drawPixmap( m_header->sectionPos( ROMViewTableHeader::PC )+17, 
                     i*lh+(lh-13)/2, *pix );    
    }
      
    // draw statistics stuff
    //
    QBrush bgBrush = (i&1) ? bgDark : bgNormal;

    // draw read count (if any)
    //
    if (m_visible[ROMViewTableHeader::READ])
    {
      if (readCnt)
      {
        const int c = QMIN( (int)(log10( (double)maxRead / 
                                         (double)readCnt ) ), 3 );

        bgBrush = (i&1) ? m_brushDark[c] : m_brush[c];

        drawInt( p, clipRect, ROMViewTableHeader::READ, y, lh, 
                 AlignLeft | AlignVCenter, readCnt, bgBrush ); 
      }
      else
      {
        bgBrush = (i&1) ? bgDark : bgNormal;

        eraseBg( p, ROMViewTableHeader::READ, y, lh, bgBrush );
      }   
    }

    // draw cycle stuff (if any)
    //
    if (avg != 0.0)
    {
      if (m_visible[ROMViewTableHeader::AVG])
      {
        drawDouble( p, clipRect, ROMViewTableHeader::AVG, y, lh, 
                    AlignLeft | AlignVCenter, avg, bgBrush );
      }

      // draw last cycles
      if (m_visible[ROMViewTableHeader::LAST])
      {
        drawInt( p, clipRect, ROMViewTableHeader::LAST, y, lh, 
                 AlignLeft | AlignVCenter, m_pic->lastCall( i ), bgBrush );
      }

      // draw min cycles
      if (m_visible[ROMViewTableHeader::MIN])
      {
        drawInt( p, clipRect, ROMViewTableHeader::MIN, y, lh, 
                 AlignLeft | AlignVCenter, m_pic->minCall( i ), bgBrush );
      }

      // draw max cycles
      if (m_visible[ROMViewTableHeader::MAX])
      {
        drawInt( p, clipRect, ROMViewTableHeader::MAX, y, lh, 
                 AlignLeft | AlignVCenter, m_pic->maxCall( i ), bgBrush );
      }
    }
    else
    {
      if (m_visible[ROMViewTableHeader::AVG])
      {
        eraseBg( p, ROMViewTableHeader::AVG,  y, lh, bgBrush );
      }
      if (m_visible[ROMViewTableHeader::LAST])
      {
        eraseBg( p, ROMViewTableHeader::LAST, y, lh, bgBrush );
      }
      if (m_visible[ROMViewTableHeader::MIN])
      {
        eraseBg( p, ROMViewTableHeader::MIN,  y, lh, bgBrush );
      }
      if (m_visible[ROMViewTableHeader::MAX])
      {
        eraseBg( p, ROMViewTableHeader::MAX,  y, lh, bgBrush );
      }
    }
    
    // draw rectangle around current line
    //
    if (m_current == i )
    {
      const int w = QMIN( visibleWidth()-2, m_header->headerWidth()-2 );
      p->drawRect( 1, i*lh+1, w, lh-2 );
    }    
  }
  
  //m_pic->unlock();
}

void
ROMViewTable::contentsMousePressEvent( QMouseEvent *ev )
{
  const int lh = lineHeight();
  const int last = m_current;
  m_current = ev->y() / lh;
  int x, y;m_popup->popup( QCursor::pos() );
  contentsToViewport( 0, last*lh, x, y );
  viewport()->update( x, y, visibleWidth(), lh ); 
  contentsToViewport( 0, m_current*lh, x, y );
  viewport()->update( x, y, visibleWidth(), lh ); 
  
  if (ev->button() == RightButton)
  {
    m_popup->clear();
    if (m_pic->rom()->isBreakpoint(m_current))
    {
      m_popup->insertItem( tr("Remove Breakpoint"), RemoveBreakpoint );
      if (m_pic->rom()->breakpointEnabled(m_current))
      {
        m_popup->insertItem( tr("Disable Breakpoint"), DeactivateBreakpoint );
      }
      else
      {
        m_popup->insertItem( tr("Enable Breakpoint"), ActivateBreakpoint );
      }
      m_popup->insertItem( tr("Edit Breakpoint"), EditBreakpoint );
    }
    else
    {
      m_popup->insertItem( tr("Set Breakpoint"), SetBreakpoint ); 
    }
    m_popup->insertSeparator();
    m_popup->insertItem( tr("Jump here"), JumpHere ); 
    m_popup->insertItem( tr("Run here"), RunHere ); 
    
    m_popup->popup( QCursor::pos() );
  }
}

void
ROMViewTable::contentsMouseDoubleClickEvent( QMouseEvent *ev )
{
  const int lh = lineHeight();
  const int last = m_current;
  m_current = ev->y() / lh;
  int x, y;
  contentsToViewport( 0, last*lh, x, y );
  viewport()->update( x, y, visibleWidth(), lh ); 
  contentsToViewport( 0, m_current*lh, x, y );
  viewport()->update( x, y, visibleWidth(), lh ); 
  
  if (ev->button() == LeftButton)
  {
    if (m_pic->rom()->isBreakpoint( m_current ))
    {
      emit showBreakpoint( m_current );
    }
  }
}

void
ROMViewTable::contentsMouseMoveEvent( QMouseEvent *ev )
{
  const int lh = lineHeight();
  const int last = m_current;
  m_current = ev->y() / lh;
  int x, y;
  contentsToViewport( 0, last*lh, x, y );
  viewport()->update( x, y, visibleWidth(), lh ); 
  contentsToViewport( 0, m_current*lh, x, y );
  viewport()->update( x, y, visibleWidth(), lh ); 
}

void
ROMViewTable::popupSLOT( int id )
{
  switch (id)
  {
  case SetBreakpoint:
    m_pic->rom()->setBreakpoint( m_current );
    emit newBreakpoint( m_current );
    break;
  case RemoveBreakpoint:
    m_pic->rom()->removeBreakpoint( m_current );
    emit removeBreakpoint( m_current );
    break;
  case EditBreakpoint:
    emit showBreakpoint( m_current );
    break;
  case ActivateBreakpoint:
    m_pic->rom()->breakpoint( m_current )->setEnabled( true );
    emit changedBreakpoint( m_current );
    break;
  case DeactivateBreakpoint:
    m_pic->rom()->breakpoint( m_current )->setEnabled( false );
    emit changedBreakpoint( m_current );
    break;
  case JumpHere:
    m_pic->setPc( m_current );
    emit refresh();
    break;
  }
  
  const int lh = lineHeight();
  int x, y;
  contentsToViewport( 0, m_current*lh, x, y );
  viewport()->update( x, y, visibleWidth(), lh ); 
}

