//======================================================================
// File:		textedit.cpp
// Author:	Matthias Toussaint
// Created:	Mon Jul 26 23:08:00 CET 2004
//----------------------------------------------------------------------
// Permission to use, copy, modify, and distribute this software and its
// documentation  for any  purpose and  without fee is  hereby  granted,
// provided  that below copyright notice appear  in all copies  and that
// both  that  copyright  notice and  this permission  notice  appear in
// supporting documentation.
// 
// This  file is  provided AS IS  with no  warranties  of any kind.  The
// author shall  have no liability  with respect  to the infringement of
// copyrights, trade  secrets  or any patents by  this file  or any part
// thereof.  In no event will the author be liable  for any lost revenue
// or profits or other special, indirect and consequential damages.
//----------------------------------------------------------------------
// (c) 2004 Matthias Toussaint
//======================================================================

#include <qpainter.h>
#include <qpixmap.h>
#include <qpopupmenu.h>
#include <qcursor.h>
#include <qlabel.h>
#include <qtooltip.h>

#include <textedit.h>
#include <inspectortip.h>
#include <pic.h>
#include <stringconversion.h>

#include <pc.xpm>
#include <bp.xpm>
#include <ibp.xpm>
#include <cbp.xpm>
#include <icbp.xpm>
#include <brownout.xpm>
#include <animation.xpm>
#include <reset.xpm>
#include <run.xpm>
#include <stop.xpm>
#include <stop_after_command.xpm>
#include <stop_after_procedure.xpm>

#include <iostream>

TextEdit::TextEdit( QWidget *parent, const char *name ) :
  QTextEdit( parent, name ),
  m_autoIndent( true ),
  m_pcLine( -1 ),
  m_followPc( true )
{
  connect( this, SIGNAL( returnPressed() ),
           this, SLOT( returnPressedSLOT() ));
  
  setFrameStyle( QFrame::WinPanel | QFrame::Sunken );
  
  setFont( QFont( "Fixed", 10, QFont::DemiBold ) );
  m_lh = fontMetrics().height()+3;
  setWordWrap( QTextEdit::NoWrap );
  
  m_margin = 40 + fontMetrics().width( "0000" );
  setMargins( m_margin, 0, 0, 0 );
  
  m_bpPix   = new QPixmap((const char **)bp_xpm);
  m_ibpPix  = new QPixmap((const char **)ibp_xpm);
  m_cbpPix  = new QPixmap((const char **)cbp_xpm);
  m_icbpPix = new QPixmap((const char **)icbp_xpm);
  m_pcPix   = new QPixmap((const char **)pc_xpm);
  
  m_brownoutPix = new QPixmap((const char **)brownout_xpm);
  m_animationPix = new QPixmap((const char **)animation_xpm);
  m_resetPix = new QPixmap((const char **)reset_xpm);
  m_runPix = new QPixmap((const char **)run_xpm);
  m_stopPix = new QPixmap((const char **)stop_xpm);
  m_stop_after_commandPix = new QPixmap((const char **)stop_after_command_xpm);
  m_stop_after_procedurePix = new QPixmap((const char **)stop_after_procedure_xpm);
  
  connect( this, SIGNAL( contentsMoving( int, int ) ),
           this, SLOT( update() ));
  
  setText( " \n" );
  sync();
  m_lh = paragraphRect(0).height();
  clear();
  
  viewport()->setMouseTracking( true );
  
  new InspectorTip( viewport() );
}

TextEdit::~TextEdit()
{
  delete m_bpPix;   
  delete m_ibpPix;  
  delete m_cbpPix;  
  delete m_icbpPix; 
  delete m_pcPix;
}

QString 
TextEdit::tokenAtPos( const QPoint & mPos )
{ 
  int para;
  const int charPos = charAt( mPos, &para );
  
  QString txt = text( para );
  
  if (!isSpace(txt[charPos]))
  {
    // maybe text
    //
    QString token;
    
    int pos = charPos;
    
    while (pos > 0 && !isSpace(txt[pos-1])) --pos;
    
    while (pos < (int)txt.length() && !isSpace(txt[pos]))
    {
      token += txt[pos++];
    }
    
    return token;
  }
  
  return QString::null;
}

int 
TextEdit::symbolAddrAtPos( const QPoint & mPos )
{ 
  return symbolAddrFromToken( tokenAtPos( mPos ) );
}

int 
TextEdit::symbolAddrFromToken( const QString & token )
{ 
  if (!token.isNull())
  {
    if (token.lower() == "w")
    {
      return -2;
    }  
    return m_pic->ram()->symbolToAddr( token );
  }
    
  return -1;
}

QRect
TextEdit::tip( const QPoint & pos, QString & text )
{
  const int addr = symbolAddrAtPos( viewportToContents( pos ) );
    
  if (addr != -1)
  {
    if (-2 == addr)
    {
      text = "<b>W</b><br><tt>Hex: 0x";
      text += StringConversion::hexString( m_pic->workingRegister() );
      text += "<br>Dec: ";
      text += StringConversion::decString( m_pic->workingRegister() );
      text += "<br>Bin: ";
      text += StringConversion::binString( m_pic->workingRegister() );
      text += "</tt>";
    }
    else
    {
      PicRegister *reg = m_pic->ram()->get( addr );

      text.sprintf( "<b>%s&nbsp;[0x%03X]</b><br><tt>Hex: 0x", 
                    reg->name().latin1(), addr );
      text += StringConversion::hexString( reg->valuePic() );
      text += "<br>Dec: ";
      text += StringConversion::decString( reg->valuePic() );
      text += "<br>Bin: ";
      text += StringConversion::binString( reg->valuePic() );
      text += "</tt>";
    }
    
    return QRect( pos.x()-10, pos.y()-10, 20, 20 );
  }
  else
  {
    text = QString::null;
  }
  
  return QRect();
}

void
TextEdit::returnPressedSLOT()
{
  if (!m_autoIndent) return;
  
  int para, index;
  getCursorPosition( &para, &index );
  QString txt = text( para-1 );
  QString ins;
  
  // txt.length()-1 because QTextEdit appends space to blank lines !?!?!?
  int pos=0;
  while (pos < ((int)txt.length())-1 && txt[pos].isSpace())
  {
    ins += txt[pos++];
  }
  insert( ins );
}

void
TextEdit::paintEvent( QPaintEvent *ev )
{
  QTextEdit::paintEvent( ev );
  
  QColor bg = colorGroup().background();
  QColor base = colorGroup().base();
  QColor textCol = colorGroup().text();
  
  QPainter p;
  p.begin(this);
  p.setClipRect( 2, 2, m_margin, height()-4 );
  
  QString addrStr;
    
  const int startLine = contentsY() / m_lh;
  const int stopLine = startLine + contentsHeight() / m_lh;
  
  PicROM *rom = m_pic->rom();
  QPixmap *pix;
  
  int lastAddr = 0;
  
  // draw breakpoints and addresses
  //
  int pos = 2+startLine*m_lh-contentsY();
  const int pixOff = m_lh/2-7;
  const int colWidth = m_margin - 4;
  const int lineX = m_margin - 2;
  
  for (int l=startLine; l<=stopLine; ++l)
  {
    const int addr = addrFromLine( l, lastAddr );
    
    // draw in stripes to avoid flicker
    //
    if (m_pcLine == l)
    {
      p.fillRect( 2, pos, colWidth, m_lh, Qt::darkGreen );
      p.fillRect( colWidth+2, pos, 4, m_lh, base );
      
      p.drawPixmap( 4, pos+m_lh/2-7, *m_pcPix );
    }
    else
    {
      p.fillRect( 2, pos, colWidth, m_lh, bg );
      p.fillRect( colWidth+2, pos, 4, m_lh, base );
    }
    
    p.drawLine( lineX, pos, lineX, pos+m_lh );
    
    if (addr > -1)
    {
      lastAddr = addr;
        
      addrStr.sprintf( "%04X", addr );
      
      if (m_pcLine == l)
      {
        p.setPen( Qt::white );
        p.drawText( 0, pos, colWidth, m_lh, 
                    Qt::AlignRight|Qt::AlignVCenter, addrStr );
        p.setPen( textCol );
      }
      else
      {
        p.drawText( 0, pos, colWidth, m_lh, 
                    Qt::AlignRight|Qt::AlignVCenter, addrStr );
      }
      
      if (rom->isBreakpoint( addr ) && !rom->temporaryBreakpoint( addr ))
      {
        if (rom->breakpointEnabled( addr ))
        {
          if (rom->breakpointConditional( addr ))
          {
            pix = m_cbpPix;
          }
          else
          {
            pix = m_bpPix;
          }
        }
        else
        {
          if (rom->breakpointConditional( addr ))
          {
            pix = m_icbpPix;
          }
          else
          {
            pix = m_ibpPix;
          }
        }
        p.drawPixmap( 20, pos+pixOff, *pix );
      }
    }
    
    pos += m_lh;
  }
    
  if (pos < height())
  {
    p.fillRect( 2, pos, colWidth, height()-pos, bg );
    p.fillRect( colWidth+2, pos, 4, height()-pos, base );
    p.drawLine( lineX, pos, lineX, height() );
  }
   
  //p.drawLine( m_margin-2, 2, m_margin-2, height()-4 );

  p.end();
}

QPopupMenu * 
TextEdit::createPopupMenu( const QPoint & pos )
{
  m_popupSymbolName = tokenAtPos( pos );
  m_popupSymbolAddr = symbolAddrFromToken( m_popupSymbolName );
  
  QPopupMenu *popup = new QPopupMenu( this );
  popup->setCheckable( true );
  
  const int lineNo = (pos.y()-1)/m_lh;
  m_popupAddr = addrFromLine( lineNo );
  
  if (m_popupAddr != -1 && !m_pic->animating())
  {
    if (m_pic->rom()->isBreakpoint( m_popupAddr ))
    {
      popup->insertItem( "Remove Breakpoint", ID_REMOVE_BREAKPOINT );
      if (m_pic->rom()->breakpointEnabled( m_popupAddr ))
      {
        popup->insertItem( "Disable Breakpoint", ID_DEACTIVATE_BREAKPOINT );
      }
      else
      {
        popup->insertItem( "Enable Breakpoint", ID_ACTIVATE_BREAKPOINT );
      }
      popup->insertSeparator();
    }
    else
    {
      popup->insertItem( "Set Breakpoint", ID_ADD_BREAKPOINT );
      popup->insertSeparator();
    }
    if (m_pic->halted())
    {
      if (m_pc != m_popupAddr)
      {
        popup->insertItem( "Jump Here", ID_JUMP_HERE );
        popup->insertItem( "Run Here", ID_RUN_HERE );
        if (m_pic->insideCall())
        {
          popup->insertItem( "Finish CALL", ID_FINISH_CALL );
        }
        popup->insertSeparator();
      }
      else if (m_pic->insideCall())
      {
        popup->insertItem( "Finish CALL", ID_FINISH_CALL );
        popup->insertSeparator();
      }
    }
  }
  else
  {
    if (m_pic->halted())
    {
      if (m_pic->insideCall() && !m_pic->animating())
      {
        popup->insertItem( "Finish CALL", ID_FINISH_CALL );
        popup->insertSeparator();
      }
    }
  }
  if (m_popupSymbolAddr != -1 && !m_pic->animating())
  {
    QString str;
    
    if (m_popupSymbolAddr >= 0)
    {
      if (m_pic->ram()->get( m_popupSymbolAddr )->watched() )
      {
        str.sprintf( tr("Remove %s from Watch"), m_popupSymbolName.latin1() );
        popup->insertItem( str, ID_REMOVE_FROM_WATCH );
      }
      else
      {
        str.sprintf( tr("Add %s to Watch"), m_popupSymbolName.latin1() );
        popup->insertItem( str, ID_ADD_TO_WATCH );
      }
      popup->insertSeparator();
    }
  }
  popup->insertItem( *m_resetPix, "Reset", ID_RESET );
  if (m_pic->halted() && !m_pic->animating())
  {
    popup->insertItem( *m_runPix, "Run", ID_RUN );
    popup->insertItem( *m_animationPix, "Animate", ID_ANIMATE );
    popup->insertItem( *m_stop_after_commandPix, "Step", ID_STEP );
    popup->insertItem( *m_stop_after_procedurePix, "Step Over", ID_STEP_OVER );
  }
  else
  {
    popup->insertItem( *m_stopPix, "Stop", ID_STOP );
  }
    
  popup->insertSeparator();
  popup->insertItem( "Follow PC", ID_FOLLOW_PC );
  popup->setItemChecked( ID_FOLLOW_PC, m_followPc );
  popup->insertItem( "Find PC", ID_FIND_PC );

  if (!m_popupSymbolName.isNull() &&
      m_pic->symbolTable()->address( m_popupSymbolName ))
  {
    m_popupSymbolAddr = 
        m_pic->symbolTable()->address( m_popupSymbolName )->value();
    QString str;
    str.sprintf( tr("Find %s"), m_popupSymbolName.latin1() );
    popup->insertItem( str, ID_FIND_ADDRESS );
  }
  
  connect( popup, SIGNAL( activated( int ) ),
           this, SLOT( popupSLOT( int ) ));
  
  return popup;
}

void
TextEdit::popupSLOT( int id )
{
  switch (id)
  {
  case ID_ADD_BREAKPOINT:
    m_pic->rom()->setBreakpoint( m_popupAddr );
    emit newBreakpointRequest( m_popupAddr );
    break;
  case ID_REMOVE_BREAKPOINT:
    m_pic->rom()->removeBreakpoint( m_popupAddr );
    emit removeBreakpointRequest( m_popupAddr );
    break;
  case ID_EDIT_BREAKPOINT:
    emit editBreakpointRequest( m_popupAddr );
    break;
  case ID_ACTIVATE_BREAKPOINT:
    m_pic->rom()->breakpoint( m_popupAddr )->setEnabled( true );
    emit changedBreakpoint( m_popupAddr );
    break;
  case ID_DEACTIVATE_BREAKPOINT:
    m_pic->rom()->breakpoint( m_popupAddr )->setEnabled( false );
    emit changedBreakpoint( m_popupAddr );
    break;
  case ID_JUMP_HERE:
    m_pic->setPc( m_popupAddr );
    setPc( m_popupAddr );
    update();
    break;
  case ID_RUN_HERE:
    emit runHereRequest( m_popupAddr );
    break;
  case ID_FINISH_CALL:
    emit finishCallRequest();
    break;
  case ID_ADD_TO_WATCH:
    emit addToWatchRequest( m_pic->ram()->get( m_popupSymbolAddr ) );
    break;
  case ID_REMOVE_FROM_WATCH:
    emit removeFromWatchRequest( m_pic->ram()->get( m_popupSymbolAddr ) );
    break;
  case ID_RESET:
    emit resetRequest();
    break;
  case ID_RUN:
    emit runRequest();
    break;
  case ID_ANIMATE:
    emit animateRequest();
    break;
  case ID_STEP:
    emit stepRequest();
    break;
  case ID_STEP_OVER:
    emit stepOverRequest();
    break;
  case ID_STOP:
    emit stopRequest();
    break;
  case ID_FOLLOW_PC:
    m_followPc = !m_followPc;
    break;
  case ID_FIND_PC:
    findPc();
    break;
  case ID_FIND_ADDRESS:
    findAddress( m_popupSymbolAddr );
    break;
  }
}

void
TextEdit::setPc( int pc )
{
  m_pc = pc;
  if (m_pc == -1) m_pc = 1;
  
  m_pcLine = m_addrToLine[m_pc]; 
  
  if (m_followPc)
  {
    findPc();
  }
}

void
TextEdit::findPc()
{
  const int pos = 1+m_pcLine*m_lh;
  
  ensureVisible( 0, pos, 0, 200 );
}

void
TextEdit::findAddress( int addr )
{
  const int pos = 1+m_addrToLine[addr]*m_lh;
  
  ensureVisible( 0, pos, 0, 200 );
}

void
TextEdit::resetPcToLineTable()
{
  for (int i=0; i<2048; ++i)
  {
    m_addrToLine[i] = -1;
  }
  
  m_pcLine = -1;
  update();
}
