//======================================================================
// File:		workspace.cpp
// Author:	Matthias Toussaint
// Created:	Mon Jul 26 23:08:00 CET 2004
//----------------------------------------------------------------------
// Permission to use, copy, modify, and distribute this software and its
// documentation  for any  purpose and  without fee is  hereby  granted,
// provided  that below copyright notice appear  in all copies  and that
// both  that  copyright  notice and  this permission  notice  appear in
// supporting documentation.
// 
// This  file is  provided AS IS  with no  warranties  of any kind.  The
// author shall  have no liability  with respect  to the infringement of
// copyrights, trade  secrets  or any patents by  this file  or any part
// thereof.  In no event will the author be liable  for any lost revenue
// or profits or other special, indirect and consequential damages.
//----------------------------------------------------------------------
// (c) 2004 Matthias Toussaint
//======================================================================

#include <qapplication.h>
#include <qdatastream.h>
#include <qfile.h>
#include <qpixmap.h>
#include <qtextedit.h>

#include <breadboardview.h>
#include <breakpointview.h>
#include <clockview.h>
#include <cpuview.h>
#include <editor.h>
#include <pic.h>
#include <ramview.h>
#include <romview.h>
#include <stackview.h>
#include <watchview.h>
#include <workspace.h>
#include <symbolview.h>
#include <instructiontraceview.h>

#include <unistd.h>

#include <breakpoint.xpm>
#include <cpu.xpm>
#include <stack.xpm>
#include <trace.xpm>
#include <symbols.xpm>
#include <watch.xpm>
#include <editor.xpm>
#include <clock.xpm>
#include <ram.xpm>
#include <rom.xpm>

#include <yapide_logo_small.xpm>


Workspace::Workspace( QWidget *parent, const char *name ) :
  QWorkspace( parent, name ),
  m_ramView( 0 ),
  m_romView( 0 ),
  m_stackView( 0 ),
  m_cpuView( 0 ),
  m_editor( 0 ),
  m_clockView( 0 ),
  m_watchView( 0 ),
  m_breakpointView( 0 ),
  m_breadboardView( 0 ),
  m_symbolView( 0 ),
  m_instructionTraceView( 0 ),
  m_pic( 0 ),
  m_animateTimer( -1 ),
  m_runUpdateTimer( -1 )
{
  m_pic = new Pic;
  
  // Gentlemen, start your engines!
  //
  m_pic->start( QThread::NormalPriority );
  m_pic->freeze();
  
  m_pic->setCrystalFrequency( 4e6 );
  //m_pic->setCrystalFrequency( 5e5 );
  
  m_yapidePix = new QPixmap((const char **)yapide_logo_small_xpm);
  
  setScrollBarsEnabled( true );
}

Workspace::~Workspace()
{
  delete m_yapidePix;
  
  m_pic->quit();
  m_pic->wait();
  
  delete m_pic;
}

void
Workspace::freeze()
{
  m_pic->freeze();
}

void
Workspace::timerEvent( QTimerEvent *ev )
{
  if (ev->timerId() == m_animateTimer)
  {
    killTimer( m_animateTimer );
    
    m_pic->step();
    while (!m_pic->halted());
    
    // expensive !!!
    //
    freeze();
    emit updateViews();
    
    if (m_pic->rom()->isActiveBreakpoint( m_pic->pc() )) stopPicSLOT();
    
    m_animateTimer = startTimer(0);
  }
  else if (ev->timerId() == m_runUpdateTimer)
  {
    killTimer( m_runUpdateTimer );
    freeze();
    emit updateViews();
    m_runUpdateTimer = startTimer( 100 );
    
    // we may have hit a breakpoint in between
    //
    if (m_pic->halted()) stopPicSLOT();
  }
}

void
Workspace::showStackSLOT( bool show )
{
  if (!m_stackView)
  {
    m_stackView = new StackView( this, "Call Stack" );
    m_stackView->setIcon( QPixmap((const char **)stack_xpm) );
    m_stackView->setCaption( tr("Stack") );
    m_stackView->setPic( m_pic );
    m_stackView->refreshSLOT();
    
    connect( this, SIGNAL( updateViews() ),
             m_stackView, SLOT( refreshSLOT() ));
  }
  
  m_stackView->setShown( show );
  /*
  BreadBoardView *board = new BreadBoardView( this );
  board->setIcon( *m_yapidePix );
  board->setCaption( tr("Breadboard") );
  connect( this, SIGNAL( updateViews() ),
           board, SLOT( refreshSLOT() ));
  board->setPic( m_pic );
  board->show();
  */
}

void
Workspace::showRAMSLOT( bool show )
{
  createRAMIfNeeded();
  
  m_ramView->setShown( show );
  m_ramView->setEditable( m_pic->halted() );
}

void
Workspace::showROMSLOT( bool show )
{
  if (!m_romView)
  {
    m_romView = new ROMView( this, "ROM" );
    m_romView->setIcon( QPixmap((const char **)rom_xpm) );
    m_romView->setCaption( tr("ROM") );
    m_romView->setPic( m_pic );
    m_romView->refreshSLOT();
    
    connect( this, SIGNAL( updateViews() ),
             m_romView, SLOT( refreshSLOT() ));
    
    connect( m_romView, SIGNAL( showBreakpoint( int ) ),
             this, SLOT( showBreakpointSLOT( int ) ));
    connect( m_romView, SIGNAL( newBreakpoint( int ) ),
             this, SLOT( newBreakpointSLOT( int ) ));
    connect( m_romView, SIGNAL( removeBreakpoint( int ) ),
             this, SLOT( removeBreakpointSLOT( int ) ));
    connect( m_romView, SIGNAL( changedBreakpoint( int ) ),
             this, SLOT( changedBreakpointSLOT( int ) ));
  }
  
  m_romView->setShown( show );
  m_romView->setEditable( m_pic->halted() );
}

void
Workspace::showCPUSLOT( bool show )
{
  if (!m_cpuView)
  {
    m_cpuView = new CPUView( this, "CPU" );
    m_cpuView->setIcon( QPixmap((const char **)cpu_xpm) );
    m_cpuView->setCaption( tr("CPU") );
    m_cpuView->setPic( m_pic );
    m_cpuView->refreshSLOT();
    
    connectRAM_CPUInfo();
    
    connect( this, SIGNAL( updateViews() ),
             m_cpuView, SLOT( refreshSLOT() ));
  }

  m_cpuView->setShown( show ); 
  m_cpuView->setEditable( m_pic->halted() );
}

void
Workspace::showClockSLOT( bool show )
{
  if (!m_clockView)
  {
    m_clockView = new ClockView( this, "Clock" );
    m_clockView->setIcon( QPixmap((const char **)clock_xpm) );
    m_clockView->setCaption( tr("Time") );
    m_clockView->setPic( m_pic );
    m_clockView->refreshSLOT();
    
    connect( this, SIGNAL( updateViews() ),
             m_clockView, SLOT( refreshSLOT() ));
  }

  m_clockView->setShown( show );
}

void
Workspace::showSymbolsSLOT( bool show )
{
  createRAMIfNeeded();
  createWatchIfNeeded();
  
  if (!m_symbolView)
  {
    m_symbolView = new SymbolView( this, "Symbols" );
    m_symbolView->setIcon( QPixmap((const char **)symbols_xpm) );
    m_symbolView->setCaption( tr("Symbols") );
    m_symbolView->setPic( m_pic );
  }
    
  connect( m_symbolView, SIGNAL( registerSelected( int ) ),
           m_ramView, SLOT( registerSelectedSLOT( int ) ));
  connect( m_symbolView, SIGNAL( addWatch( PicRegister * ) ),
           m_watchView, SLOT( addWatchSLOT( PicRegister * ) ));
  connect( m_symbolView, SIGNAL( removeWatch( PicRegister * ) ),
           m_watchView, SLOT( removeWatchSLOT( PicRegister * ) ));

  m_symbolView->setShown( show );
}

void
Workspace::showWatchSLOT( bool show )
{
  createWatchIfNeeded();
  
  m_watchView->setShown( show );
  m_watchView->setEditable( m_pic->halted() );
}

void
Workspace::showInstructionTraceSLOT( bool show )
{
  if (!m_instructionTraceView)
  {
    m_instructionTraceView = new InstructionTraceView( this, "InstructionTrace" );
    m_instructionTraceView->setIcon( QPixmap((const char **)trace_xpm) );
    m_instructionTraceView->setCaption( tr("Instruction Trace") );
    m_instructionTraceView->setPic( m_pic );
    
    connect( this, SIGNAL( updateViews() ),
             m_instructionTraceView, SLOT( refreshSLOT() ));
    
    if (m_editor)
    {
      connect( m_instructionTraceView, SIGNAL( showPc( int ) ),
               m_editor, SLOT( gotoPcSLOT( int ) ));
    }
  }
  
  m_instructionTraceView->setShown( show );
}

void
Workspace::showBreakpointsSLOT( bool show )
{
  createBreakpointsIfNeeded();
  
  m_breakpointView->setShown( show );
  m_breakpointView->setEditable( m_pic->halted() );
}

void
Workspace::createWatchIfNeeded()
{
  if (!m_watchView)
  {
    m_watchView = new WatchView( this, "Watch" );
    m_watchView->setIcon( QPixmap((const char **)watch_xpm) );
    m_watchView->setCaption( tr("Watch") );
    m_watchView->setPic( m_pic );
    m_watchView->refreshSLOT();
    
    connectRAM_Watch();
            
    connect( this, SIGNAL( updateViews() ),
             m_watchView, SLOT( refreshSLOT() ));
        
    m_watchView->hide();
  }
}

void
Workspace::createRAMIfNeeded()
{
  if (!m_ramView)
  {
    m_ramView = new RAMView( this, "RAM" );
    m_ramView->setIcon( QPixmap((const char **)ram_xpm) );
    m_ramView->setCaption( tr("RAM") );
    m_ramView->setPic( m_pic );
    m_ramView->refreshSLOT();
    
    connect( this, SIGNAL( updateViews() ),
             m_ramView, SLOT( refreshSLOT() ));
    
    connect( m_ramView, SIGNAL( addTrace( PicRegister * ) ),
             this, SLOT( addTraceSLOT( PicRegister * )));
    connect( m_ramView, SIGNAL( removeTrace( PicRegister * ) ),
             this, SLOT( removeTraceSLOT( PicRegister * )));
    
    connectRAM_Watch();
    connectRAM_CPUInfo();  
    
    m_ramView->hide();
  }
}

void
Workspace::createBreakpointsIfNeeded()
{
  if (!m_breakpointView)
  {
    m_breakpointView = new BreakpointView( this, "Breakpoints" );
    m_breakpointView->setIcon( QPixmap((const char **)breakpoint_xpm) );
    m_breakpointView->setCaption( tr("Breakpoints") );
    m_breakpointView->setPic( m_pic );
    
    m_breakpointView->hide();
    
    connect( m_breakpointView, SIGNAL( changedBreakpoint( int ) ),
             this, SLOT( changedBreakpointSLOT( int ) ));
  }
}

void
Workspace::showEditorSLOT( bool show )
{
  if (!m_editor)
  {
    m_editor = new Editor( this );
    m_editor->setIcon( QPixmap((const char **)editor_xpm) );
    m_editor->setPic( m_pic );
    m_editor->setCaption( tr("Editor") );
    
    connect( this, SIGNAL( updateViews() ),
             m_editor, SLOT( refreshSLOT() ));
    connect( m_editor, SIGNAL( reloadPic() ),
             this, SLOT( reloadPicSLOT() ));
    
    connect( m_editor, SIGNAL( runRequest() ),
             this, SLOT( runPicSLOT() ));
    connect( m_editor, SIGNAL( finishCallRequest() ),
             this, SLOT( finishCallPicSLOT() ));
    connect( m_editor, SIGNAL( animateRequest() ),
             this, SLOT( animatePicSLOT() ));
    connect( m_editor, SIGNAL( stopRequest() ),
             this, SLOT( stopPicSLOT() ));
    connect( m_editor, SIGNAL( stepRequest() ),
             this, SLOT( stepPicSLOT() ));
    connect( m_editor, SIGNAL( stepOverRequest() ),
             this, SLOT( stepOverPicSLOT() ));
    
    connect( m_editor, SIGNAL( addToWatchRequest( PicRegister * ) ),
             m_watchView, SLOT( addWatchSLOT( PicRegister * ) ));
    connect( m_editor, SIGNAL( removeFromWatchRequest( PicRegister * ) ),
             m_watchView, SLOT( removeWatchSLOT( PicRegister * ) ));
    
    connect( m_editor, SIGNAL( showBreakpointRequest( int ) ),
             this, SLOT( showBreakpointSLOT( int ) ));
    connect( m_editor, SIGNAL( newBreakpointRequest( int ) ),
             this, SLOT( newBreakpointSLOT( int ) ));
    connect( m_editor, SIGNAL( removeBreakpointRequest( int ) ),
             this, SLOT( removeBreakpointSLOT( int ) ));

    connect( m_editor, SIGNAL( changedBreakpoint( int ) ),
             this, SLOT( changedBreakpointSLOT( int ) ));
    connect( m_editor, SIGNAL( message( const QString & ) ),
             this, SIGNAL( message( const QString & ) ));
    
    connect( m_editor, SIGNAL( updateViews() ),
             this, SIGNAL( updateViews() ));
    
    if (m_instructionTraceView)
    {
      connect( m_instructionTraceView, SIGNAL( showPc( int ) ),
               m_editor, SLOT( gotoPcSLOT( int ) ));
    }
  }
  
  m_editor->setShown( show );
}
  
void
Workspace::connectRAM_Watch()
{  
  if (!m_watchView)
  {
    createWatchIfNeeded();
  }
  else if (m_ramView) 
  {
    connect( m_ramView, SIGNAL( addWatch( PicRegister * ) ),
             m_watchView, SLOT( addWatchSLOT( PicRegister * ) ));
    connect( m_ramView, SIGNAL( removeWatch( PicRegister * ) ),
             m_watchView, SLOT( removeWatchSLOT( PicRegister * ) ));
    connect( m_watchView, SIGNAL( updateViews() ),
             this, SIGNAL( updateViews() ));
    connect( m_ramView, SIGNAL( updateViews() ),
             this, SIGNAL( updateViews() ));
  }
}

void
Workspace::connectRAM_CPUInfo()
{
  if (m_cpuView && m_ramView) 
  {
    connect( m_cpuView, SIGNAL( registerChanged( PicRegister * ) ),
             m_ramView, SLOT( registerChangedSLOT( PicRegister * ) ));
  }
}
  
void 
Workspace::resetPicSLOT()
{
  m_pic->reset();
  while (m_pic->rstRequest());
  
  // if nothing helps -> sleep :o(
  //
  usleep( 10000 );
  
  freeze();
  emit updateViews();  
}

void 
Workspace::runPicSLOT()
{ 
  if (m_pic->dirty())
  {
    qApp->beep();
  }
  else
  {
    emit updateMenus( true );

    setEditable( false );

    m_runUpdateTimer = startTimer( 100 );

    m_pic->cont();
  }
}

void 
Workspace::stepPicSLOT()
{
  if (m_pic->dirty())
  {
    qApp->beep();
  }
  else
  {
    m_pic->step();
    while (!m_pic->halted());

    freeze();
    emit updateViews();  
    emit updateMenus( false );
  }
}

void 
Workspace::overPicSLOT()
{
  if (m_pic->dirty())
  {
    qApp->beep();
  }
  else
  {
    m_pic->over();
    emit updateMenus( true );
    setEditable( false );
    
    // if nothiong helps -> sleep :o(
    //
    usleep( 1000 );
    
    m_runUpdateTimer = startTimer( 0 );
  }
}

void 
Workspace::finishCallPicSLOT()
{
  if (m_pic->dirty())
  {
    qApp->beep();
  }
  else
  {
    m_pic->finishCall();
    emit updateMenus( true );
    setEditable( false );
    
    // if nothiong helps -> sleep :o(
    //
    usleep( 1000 );
    
    m_runUpdateTimer = startTimer( 0 );
  }
}

void 
Workspace::stopPicSLOT()
{
  m_pic->setAnimating( false );
  
  if (m_runUpdateTimer != -1)
  {
    killTimer( m_runUpdateTimer );
    m_runUpdateTimer = -1;
  }
  
  if (m_animateTimer != -1)
  {
    killTimer( m_animateTimer );
    m_animateTimer = -1;
  }
  
  if (!m_pic->halted()) m_pic->halt();
  while (!m_pic->halted());
  
  emit updateMenus( false );
  freeze();
  emit updateViews(); 
  
  setEditable( true );
}

void 
Workspace::animatePicSLOT()
{
  if (m_pic->dirty())
  {
    qApp->beep();
  }
  else
  {
    if (m_animateTimer == -1)
    {
      m_animateTimer = startTimer(0);
      m_pic->setAnimating( true );
    }
    else
    {
      killTimer( m_animateTimer );
      m_animateTimer = -1;
      m_pic->setAnimating( false );
    }

    emit updateMenus( false );
  }
}

void 
Workspace::refreshSLOT()
{
  emit updateViews();
}

void 
Workspace::setEditable( bool editable )
{
  if (m_cpuView) m_cpuView->setEditable( editable );
  if (m_ramView) m_ramView->setEditable( editable );
  if (m_romView) m_romView->setEditable( editable );
  if (m_watchView) m_watchView->setEditable( editable );
}

void
Workspace::showBreakpointSLOT( int )
{
  showBreakpointsSLOT( true );
}

void
Workspace::newBreakpointSLOT( int addr )
{
  createBreakpointsIfNeeded();
  
  if (m_editor)
  {
    m_editor->updateBreakpoints();
  }
  if (m_romView)
  {
    m_romView->refreshSLOT();
  }
  
  m_breakpointView->newBreakpointSLOT( addr );
}

void
Workspace::removeBreakpointSLOT( int addr )
{ 
  createBreakpointsIfNeeded();
  
  if (m_editor)
  {
    m_editor->updateBreakpoints();
  }
  if (m_romView)
  {
    m_romView->refreshSLOT();
  }
  
  m_breakpointView->removeBreakpointSLOT( addr );
}

void
Workspace::changedBreakpointSLOT( int addr )
{
  createBreakpointsIfNeeded();
  
  if (m_editor)
  {
    m_editor->updateBreakpoints();
  }
  if (m_romView)
  {
    m_romView->refreshSLOT();
  }
  
  m_breakpointView->changedBreakpointSLOT( addr );
}

void 
Workspace::showFindSLOT()
{
  if (m_editor)
  {
    m_editor->showFindSLOT();
  }
}

void 
Workspace::findAgainSLOT()
{
  if (m_editor)
  {
    m_editor->findAgainSLOT();
  }
}

void 
Workspace::showReplaceSLOT()
{
  if (m_editor)
  {
    m_editor->showReplaceSLOT();
  }
}

void 
Workspace::replaceAgainSLOT()
{
  if (m_editor)
  {
    m_editor->replaceAgainSLOT();
  }
}

void 
Workspace::gotoLineSLOT()
{
  if (m_editor)
  {
    m_editor->gotoLineSLOT();
  }
}

void 
Workspace::gotoSelectionSLOT()
{
  if (m_editor)
  {
    m_editor->gotoSelectionSLOT();
  }
}

void 
Workspace::cutSLOT()
{
  if (m_editor)
  {
    m_editor->cutSLOT();
  }
}

void 
Workspace::copySLOT()
{
  if (m_editor)
  {
    m_editor->copySLOT();
  }
}

void 
Workspace::pasteSLOT()
{
  if (m_editor)
  {
    m_editor->pasteSLOT();
  }
}

void 
Workspace::selectAllSLOT()
{
  if (m_editor)
  {
    m_editor->selectAllSLOT();
  }
}

void 
Workspace::undoSLOT()
{
  if (m_editor)
  {
    m_editor->undoSLOT();
  }
}

void 
Workspace::redoSLOT()
{
  if (m_editor)
  {
    m_editor->redoSLOT();
  }
}

void 
Workspace::makeSLOT()
{
  if (m_editor)
  {
    m_editor->makeSLOT();
  }
}

void 
Workspace::saveSLOT()
{
  if (m_editor)
  {
    m_editor->saveSLOT();
  }
}

void 
Workspace::openSLOT()
{
  showEditorSLOT( true );
  
  m_editor->openSLOT();
}

void 
Workspace::upperSLOT()
{
  if (m_editor)
  {
    m_editor->upperSLOT();
  }
}

void 
Workspace::lowerSLOT()
{
  if (m_editor)
  {
    m_editor->lowerSLOT();
  }
}

void 
Workspace::freeRunSLOT( bool on )
{
  m_pic->setFreeRun( on );
}

void 
Workspace::tracingSLOT( bool on )
{
  m_pic->setTracing( on );
}

void
Workspace::reloadPicSLOT()
{
  if (m_ramView) m_ramView->setPic( m_pic );
  if (m_romView) m_romView->setPic( m_pic );
  if (m_cpuView) m_cpuView->setPic( m_pic );
  if (m_symbolView) m_symbolView->setPic( m_pic );
}

void
Workspace::addTraceSLOT( PicRegister *reg )
{
  m_pic->registerTrace()->addRegister( reg );
}

void
Workspace::removeTraceSLOT( PicRegister *reg )
{
  m_pic->registerTrace()->removeRegister( reg );
}
